/*
 * Session Caching Daemon
 *
 * Copyright (C) 2008 Juha Railoma <jrma@iki.fi>
 *
 *
 * This file is part of Session Caching Daemon.
 *
 * Session Caching Daemon is free software: you can redistribute
 * it and/or modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later version.
 *
 * Session Caching Daemon is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Session Caching Daemon.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *
 * For license terms, see the file COPYING in this directory.
 */

// TODO: prepare session-header yhteen emallociin
// TODO: php_info() - openconns
// TODO: dokumentoinit + sematiikka
// TODO: jrkevt paluuarvot
// TODO: zts-testi
// TODO: scache_lasterror

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <assert.h>

#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/uio.h>

#include <sys/socket.h>
#include <sys/un.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>

#include <poll.h>

#include <php.h>
#include <php_ini.h>
#include <php_network.h>
#include <zend_constants.h>
#include <zend_exceptions.h>
#include <ext/standard/info.h>
#include <ext/standard/php_var.h>
#include <ext/standard/php_smart_str.h>
#include <ext/session/php_session.h>

#include "php-scache-proto.h"
#include "php-scache-paths.h"

/* from config.h, enough functionality to enable ipv6 */
#if defined(HAVE_GETADDRINFO) && \
    defined(HAVE_FREEADDRINFO) && \
    defined(HAVE_NETINET_IP6_H) && \
    defined(HAVE_SOCKADDR_STORAGE) && \
	!defined(DISABLE_IPV6)

# include "netinet/ip6.h"
# define ENABLE_IPV6 1
#endif


#ifdef ZTS
# include "TSRM.h"

# define SCACHE_G(v) TSRMG(scache_globals_id, zend_scache_globals *, v)
#else
# define SCACHE_G(v) (scache_globals.v)
#endif

/* sockets and ports */
#if defined(ENABLE_DEVEL)
# define SCACHE_PTO_UNIX_SOCK       "/tmp/scached-devel.sock"
#elif defined(SCACHED_SOCK)
# define SCACHE_PTO_UNIX_SOCK       SCACHED_SOCK
#else 
# define SCACHE_PTO_UNIX_SOCK       "/tmp/scached.sock"
#endif

#if defined(SCACHED_PORT)
# define SCACHE_PTO_LISTEN_PORT      SCACHED_PORT
#else
# define SCACHE_PTO_LISTEN_PORT      50659
#endif

/* need-more-than-off-by-one */
#define _NMTOBO            16

/* max iov if not resolvable */
#define DEFAULT_MAX_IOV      16
#define DEFAULT_CONN_TIMEOUT 5000

/* resource name for connection */
#define SC_SESS_RESNAME    "scache connection"
#define M() printf("MARK: %s:%u\n", __FUNCTION__, __LINE__);

/* array-size */
#define sc_asize(x)  (sizeof(x) / sizeof(x[0]))

/* local structures */
struct sc_conn_t {

	int fd;
	int dirty; /* set on io, clean on success, if dirty when
				  starting, reconnect */
	int persistent;
	int port;
	char *host;
	struct sc_conn_t *next; /* persistents are queued */
};

struct sc_wblock_t {

	int pos;
	int length;
	char *base;

	struct sc_wblock_t *next;
};


struct sc_session_t {
	
	int index;

	char *tag;
	char *secret;

	enum scache_error_t scerr;
	struct sc_wblock_t *header;
	struct sc_conn_t *conn;
};

struct sc_req_t {

	enum scache_reqtype_t reqtype;

	int flags;
	int rlen;
	char *rbuffer;

	struct sc_wblock_t *wblocks;
};

struct sc_iop_t {

	long iop;
	
	char *arg1;
	char *arg2;

	int arg2_len;
	char arg2_prefix;

	zval *rval;
	enum scache_error_t rerr;

	struct sc_iop_t *next;
};

struct sc_classdata_t {
	zend_object std; /* required?? */
	long expire;
	struct sc_session_t *session;
};


/* +++++++++++++++++++++++++++++++++++++++++++++++++++
 * Local variables
 */


static int le_sessions;
static int max_iov               = DEFAULT_MAX_IOV;
static int max_retries           = 2;
static int max_connect_timeout   = 5000;

static int persist_on            = 1;
static int persist_off           = 0;

static const char *scache_errstr[] = {
    "SCERR_SUCCESS",             /* 0 */
    "SCERR_PROTOCOL",            /* 1 */
    "SCERR_EXIST",               /* 2 */
    "SCERR_NOTEXIST",            /* 3 */
    "SCERR_SYSTEM",              /* 4 */
    "SCERR_INVALID_REQUEST",     /* 5 */
    "SCERR_UNKNOWN_REQUEST",     /* 6 */
    "SCERR_LIMITS_REACHED",      /* 7 */
    "SCERR_NOT_CONNECTED",       /* 8 */
    "SCERR_NO_SESSION",          /* 9 */
	"SCERR_NO_ACCESS",           /* 10 */
	"SCERR_BREAK",               /* 11 */
	"SCERR_INVALID_FORMAT"       /* 12 */
};

/* class */
zend_class_entry *scache_ce_scache;
zend_class_entry *scache_ce_exception;

/* +++++++++++++++++++++++++++++++++++++++++++++++++++
 * PHP-stuff
 */

ZEND_BEGIN_MODULE_GLOBALS(scache)
int   ini_default_port;
int   ini_conn_timeout;
char *ini_default_host;
char *ini_tag_suffix;
char *ini_auth_string;
struct sc_conn_t *sc_conn_list;
ZEND_END_MODULE_GLOBALS(scache)

ZEND_DECLARE_MODULE_GLOBALS(scache)


ZEND_MINIT_FUNCTION(scache);
ZEND_MSHUTDOWN_FUNCTION(scache);
ZEND_RINIT_FUNCTION(scache);
ZEND_RSHUTDOWN_FUNCTION(scache);
ZEND_MINFO_FUNCTION(scache);


/* FUNCTIONS */
ZEND_FUNCTION(scache_open);
ZEND_FUNCTION(scache_popen);
ZEND_FUNCTION(scache_reset);
ZEND_FUNCTION(scache_preset);
ZEND_FUNCTION(scache_connect);
ZEND_FUNCTION(scache_pconnect);
ZEND_FUNCTION(scache_connect_fast);
ZEND_FUNCTION(scache_pconnect_fast);
ZEND_FUNCTION(scache_rename);
ZEND_FUNCTION(scache_drop);

ZEND_FUNCTION(scache_getopt);
ZEND_FUNCTION(scache_setopt);
ZEND_FUNCTION(scache_status);
ZEND_FUNCTION(scache_srvstats);
ZEND_FUNCTION(scache_getindex);
ZEND_FUNCTION(scache_lasterr);
ZEND_FUNCTION(scache_strerror);

ZEND_FUNCTION(scache_iov);

ZEND_FUNCTION(scache_get);
ZEND_FUNCTION(scache_unset);
ZEND_FUNCTION(scache_stat);
ZEND_FUNCTION(scache_exists);
ZEND_FUNCTION(scache_set);
ZEND_FUNCTION(scache_add);
ZEND_FUNCTION(scache_replace);
ZEND_FUNCTION(scache_coll);

ZEND_FUNCTION(scache_shget);
ZEND_FUNCTION(scache_shunset);
ZEND_FUNCTION(scache_shstat);
ZEND_FUNCTION(scache_shexists);
ZEND_FUNCTION(scache_shset);
ZEND_FUNCTION(scache_shadd);
ZEND_FUNCTION(scache_shreplace);
ZEND_FUNCTION(scache_shcoll);

ZEND_FUNCTION(scache_chput);
ZEND_FUNCTION(scache_chget);
ZEND_FUNCTION(scache_chclear);

ZEND_FUNCTION(scache_vadd);
ZEND_FUNCTION(scache_vsub);
ZEND_FUNCTION(scache_vget);
ZEND_FUNCTION(scache_vset);
ZEND_FUNCTION(scache_vunset);

ZEND_FUNCTION(scache_rnpush);
ZEND_FUNCTION(scache_rnpop);
ZEND_FUNCTION(scache_rnshift);
ZEND_FUNCTION(scache_rnunshift);
ZEND_FUNCTION(scache_rnget);
ZEND_FUNCTION(scache_rnrotf);
ZEND_FUNCTION(scache_rnrotb);
ZEND_FUNCTION(scache_rnsize);
ZEND_FUNCTION(scache_rnclear);

zend_function_entry scache_functions[] = {
	ZEND_FE(scache_open, NULL)
	ZEND_FE(scache_popen, NULL)
	ZEND_FE(scache_reset, NULL)
	ZEND_FE(scache_preset, NULL)
	ZEND_FE(scache_connect, NULL)
	ZEND_FE(scache_pconnect, NULL)
	ZEND_FE(scache_connect_fast, NULL)
	ZEND_FE(scache_pconnect_fast, NULL)
	ZEND_FE(scache_rename, NULL)
	ZEND_FE(scache_drop, NULL)

	ZEND_FE(scache_getopt, NULL)
	ZEND_FE(scache_setopt, NULL)
	ZEND_FE(scache_status, NULL)
	ZEND_FE(scache_srvstats, NULL)
	ZEND_FE(scache_getindex, NULL)
	ZEND_FE(scache_lasterr, NULL)
	ZEND_FE(scache_strerror, NULL)
	ZEND_FE(scache_iov, NULL)

	ZEND_FE(scache_get, NULL)
	ZEND_FE(scache_unset, NULL)
	ZEND_FE(scache_stat, NULL)
	ZEND_FE(scache_exists, NULL)
	ZEND_FE(scache_set, NULL)
	ZEND_FE(scache_add, NULL)
	ZEND_FE(scache_replace, NULL)
	ZEND_FE(scache_coll, NULL)

	ZEND_FE(scache_shget, NULL)
	ZEND_FE(scache_shunset, NULL)
	ZEND_FE(scache_shstat, NULL)
	ZEND_FE(scache_shexists, NULL)
	ZEND_FE(scache_shset, NULL)
	ZEND_FE(scache_shadd, NULL)
	ZEND_FE(scache_shreplace, NULL)
	ZEND_FE(scache_shcoll, NULL)

	ZEND_FE(scache_chget, NULL)
	ZEND_FE(scache_chput, NULL)
	ZEND_FE(scache_chclear, NULL)

	ZEND_FE(scache_vadd, NULL)
	ZEND_FE(scache_vsub, NULL)
	ZEND_FE(scache_vget, NULL)
	ZEND_FE(scache_vset, NULL)
	ZEND_FE(scache_vunset, NULL)

	ZEND_FE(scache_rnpush, NULL)
	ZEND_FE(scache_rnpop, NULL)
	ZEND_FE(scache_rnshift, NULL)
	ZEND_FE(scache_rnunshift, NULL)
	ZEND_FE(scache_rnget, NULL)
	ZEND_FE(scache_rnrotf, NULL)
	ZEND_FE(scache_rnrotb, NULL)
	ZEND_FE(scache_rnsize, NULL)
	ZEND_FE(scache_rnclear, NULL)
	
	ZEND_FALIAS(scache_clear, scache_unset, NULL)
	ZEND_FALIAS(scache_shclear, scache_shunset, NULL)
	ZEND_FALIAS(scache_vclear, scache_vunset, NULL)

	{NULL, NULL, NULL}
};

/* class-interface */
ZEND_METHOD(SCache, __construct);
ZEND_METHOD(SCache, open);
ZEND_METHOD(SCache, reset);
ZEND_METHOD(SCache, rename);
ZEND_METHOD(SCache, drop);

ZEND_METHOD(SCache, getopt);
ZEND_METHOD(SCache, setopt);
ZEND_METHOD(SCache, status);
ZEND_METHOD(SCache, srvstats);
ZEND_METHOD(SCache, getindex);
ZEND_METHOD(SCache, lasterr);
ZEND_METHOD(SCache, strerror);

ZEND_METHOD(SCache, iov);

ZEND_METHOD(SCache, get);
ZEND_METHOD(SCache, clear);
ZEND_METHOD(SCache, stat);
ZEND_METHOD(SCache, exists);
ZEND_METHOD(SCache, set);
ZEND_METHOD(SCache, add);
ZEND_METHOD(SCache, replace);
ZEND_METHOD(SCache, coll);

ZEND_METHOD(SCache, shget);
ZEND_METHOD(SCache, shclear);
ZEND_METHOD(SCache, shstat);
ZEND_METHOD(SCache, shexists);
ZEND_METHOD(SCache, shset);
ZEND_METHOD(SCache, shadd);
ZEND_METHOD(SCache, shreplace);
ZEND_METHOD(SCache, shcoll);

ZEND_METHOD(SCache, chget);
ZEND_METHOD(SCache, chput);
ZEND_METHOD(SCache, chclear);

ZEND_METHOD(SCache, vadd);
ZEND_METHOD(SCache, vsub);
ZEND_METHOD(SCache, vget);
ZEND_METHOD(SCache, vset);
ZEND_METHOD(SCache, vclear);

ZEND_METHOD(SCache, rnpush);
ZEND_METHOD(SCache, rnpop);
ZEND_METHOD(SCache, rnshift);
ZEND_METHOD(SCache, rnunshift);
ZEND_METHOD(SCache, rnget);
ZEND_METHOD(SCache, rnrotf);
ZEND_METHOD(SCache, rnrotb);
ZEND_METHOD(SCache, rnsize);
ZEND_METHOD(SCache, rnclear);

zend_function_entry scache_methods[] = {
	ZEND_ME(SCache, __construct, NULL, ZEND_ACC_PUBLIC | ZEND_ACC_CTOR)

	ZEND_ME(SCache, open, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, reset, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, rename, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, drop, NULL, ZEND_ACC_PUBLIC)

	ZEND_ME(SCache, getopt, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, setopt, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, status, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, srvstats, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, getindex, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, lasterr, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, strerror, NULL, ZEND_ACC_PUBLIC)

	ZEND_ME(SCache, iov, NULL, ZEND_ACC_PUBLIC)

	ZEND_ME(SCache, get, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, clear, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, stat, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, exists, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, set, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, add, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, replace, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, coll, NULL, ZEND_ACC_PUBLIC)
	
	ZEND_ME(SCache, shget, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, shclear, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, shstat, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, shexists, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, shset, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, shadd, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, shreplace, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, shcoll, NULL, ZEND_ACC_PUBLIC)
	
	ZEND_ME(SCache, chget, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, chput, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, chclear, NULL, ZEND_ACC_PUBLIC)
	
	ZEND_ME(SCache, vadd, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, vsub, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, vget, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, vset, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, vclear, NULL, ZEND_ACC_PUBLIC)

	ZEND_ME(SCache, rnpush, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, rnpop, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, rnshift, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, rnunshift, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, rnget, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, rnrotf, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, rnrotb, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, rnsize, NULL, ZEND_ACC_PUBLIC)
	ZEND_ME(SCache, rnclear, NULL, ZEND_ACC_PUBLIC)

	{NULL, NULL, NULL}
};

zend_module_entry scache_module_entry = {
#if ZEND_MODULE_API_NO >= 20020306 // (sessions_api > 20020306) 20010901
	STANDARD_MODULE_HEADER,
#endif
	"scache",
	scache_functions,
	ZEND_MINIT(scache),
	ZEND_MSHUTDOWN(scache),
	ZEND_RINIT(scache),
	ZEND_RSHUTDOWN(scache),
	ZEND_MINFO(scache),
#if ZEND_MODULE_API_NO >= 20010901
	NO_VERSION_YET,
#endif
	STANDARD_MODULE_PROPERTIES
};

#ifdef COMPILE_DL_SCACHE
ZEND_GET_MODULE(scache)
#endif



/* session save handler, if sessions available */
#ifdef HAVE_PHP_SESSION
PS_FUNCS(scache);
ps_module ps_mod_scache = {
	PS_MOD(scache)
};

#endif /* HAVE_PHP_SESSION */


/* ==============================
 * ini-values 
 * ============================== */
#define SCINI_MIN_DEFAULT_PORT   0
#define SCINI_MAX_DEFAULT_PORT   65535
#define SCINI_MIN_CONN_TIMEOUT   0
#define SCINI_MAX_CONN_TIMEOUT   5000
#define SCINI_MAX_TAG_SUFFIX_LEN 63

#define SCINI_TAG_SUFFIX        "scache.tag_suffix"
#define SCINI_AUTH_STRING       "scache.auth_string"
#define SCINI_CONN_TIMEOUT      "scache.connect_timeout"
#define SCINI_DEFAULT_HOST      "scache.default_host"
#define SCINI_DEFAULT_PORT      "scache.default_port"

PHP_INI_MH(sc_ini_tag_suffix_changed);
PHP_INI_MH(sc_ini_auth_string_changed);
PHP_INI_MH(sc_ini_default_host_changed);
PHP_INI_MH(sc_ini_default_port_changed);
PHP_INI_MH(sc_ini_conn_timeout_changed);

PHP_INI_BEGIN()
	PHP_INI_ENTRY(SCINI_TAG_SUFFIX, NULL, PHP_INI_SYSTEM, 
				  sc_ini_tag_suffix_changed)
	PHP_INI_ENTRY(SCINI_AUTH_STRING, NULL, PHP_INI_PERDIR,
				  sc_ini_auth_string_changed)
	PHP_INI_ENTRY(SCINI_CONN_TIMEOUT, NULL, PHP_INI_SYSTEM, 
				  sc_ini_conn_timeout_changed)
	PHP_INI_ENTRY(SCINI_DEFAULT_HOST, NULL, PHP_INI_ALL, 
		          sc_ini_default_host_changed)
	PHP_INI_ENTRY(SCINI_DEFAULT_PORT, NULL, PHP_INI_ALL,
				  sc_ini_default_port_changed)
PHP_INI_END()

/* for debugging, defaults to native */
#if (0)
static void *
sc_pecalloc(int size, int num, int persistent) {
	void *n = pecalloc(size, num, persistent);
	printf("sc_pecalloc(%p)\n", n);
	return n;
}

static void *
sc_pemalloc(int size, int persistent) {
	void *n = pemalloc(size, persistent);
	printf("sc_pemalloc(%p)\n", n);
	return n;
}

static void *
sc_emalloc(int size) {
	void *n = emalloc(size);
	printf("sc_emalloc(%p)\n", n);
	return n;
}

static char *
sc_pestrdup(const char *src, int persistent) {
	char *n = pestrdup(src, 1);
	printf("sc_pestrdup(%p)\n", n);
	return n;
}

static void
sc_pefree(void *addr, int persistent) {
	printf("sc_pefree(%p)\n", addr);
	return pefree(addr, persistent);
}

static void
sc_efree(void *addr) {
	printf("sc_efree(%p)\n", addr);
	return efree(addr);
}
#else
# define sc_pecalloc pecalloc
# define sc_pemalloc pemalloc
# define sc_emalloc  emalloc
# define sc_pestrdup pestrdup
# define sc_pefree   pefree
# define sc_efree    efree
#endif

static void 
php_scache_init_globals(zend_scache_globals *scache_globals)
{

	/* connection-queue */
	scache_globals->sc_conn_list = NULL;
	scache_globals->ini_default_host = NULL;
	scache_globals->ini_default_port = 0;
	scache_globals->ini_conn_timeout = DEFAULT_CONN_TIMEOUT;
	scache_globals->ini_tag_suffix = sc_pestrdup("", 1);
	scache_globals->ini_auth_string = sc_pestrdup("", 1);
}

static void
sc_ini_tag_suffix_changed_i() {

	char *cp;

	cp = SCACHE_G(ini_tag_suffix);
	if (cp) sc_pefree(cp, 1);
	SCACHE_G(ini_tag_suffix) = NULL;

	cp = INI_STR(SCINI_TAG_SUFFIX);
	if (cp) SCACHE_G(ini_tag_suffix) = sc_pestrdup(cp, 1);
}

PHP_INI_MH(sc_ini_tag_suffix_changed) {
	sc_ini_tag_suffix_changed_i();
	return SUCCESS;
}

static void
sc_ini_auth_string_changed_i() {

	char *cp;

	/* free previous */
	cp = SCACHE_G(ini_auth_string);
	if (cp) sc_pefree(cp, 1);
	SCACHE_G(ini_auth_string) = NULL;

	/* get new if valid */
	cp = INI_STR(SCINI_AUTH_STRING);
	if (cp) {

		if ((strlen(cp) < SCACHE_PTO_MIN_AUTHLEN) ||
			(strlen(cp) > SCACHE_PTO_MAX_AUTHLEN)) {

			zend_error(E_WARNING, "auth string in %s too long",
					   SCINI_AUTH_STRING);

		} else {
			SCACHE_G(ini_auth_string) = sc_pestrdup(cp, 1);
		}
	}
}

PHP_INI_MH(sc_ini_auth_string_changed) {
	sc_ini_auth_string_changed_i();
	return SUCCESS;
}

static void
sc_ini_default_host_changed_i() {

	char *cp;

	/* release old, and strdup new */
	cp = SCACHE_G(ini_default_host);
	if (cp) sc_pefree(cp, 1);

	cp = INI_STR(SCINI_DEFAULT_HOST);
	if (cp) SCACHE_G(ini_default_host) = sc_pestrdup(cp, 1);
	else SCACHE_G(ini_default_host) = NULL;
}

PHP_INI_MH(sc_ini_default_host_changed) {
	sc_ini_default_host_changed_i();
	return SUCCESS;
}

static void
sc_ini_default_port_changed_i() {

	int i = INI_INT(SCINI_DEFAULT_PORT);

	/* value 0 == SCACHE_PTO_LISTEN_PORT */
	if ((i < SCINI_MIN_DEFAULT_PORT) ||
		(i > SCINI_MAX_DEFAULT_PORT)) {

		zend_error(E_WARNING, "port value in %s out of bounds",
				   SCINI_DEFAULT_PORT);
	} else {
		SCACHE_G(ini_default_port) = i;
	}
}

PHP_INI_MH(sc_ini_default_port_changed) {
	sc_ini_default_port_changed_i();
	return SUCCESS;
}

static void
sc_ini_conn_timeout_changed_i() {

	int i = INI_INT(SCINI_CONN_TIMEOUT);

	/* value 0 == DEFAULT_CONN_TIMEOUT */
	if ((i < SCINI_MIN_CONN_TIMEOUT) ||
		(i > SCINI_MAX_CONN_TIMEOUT)) {

		zend_error(E_WARNING, "timeout value in %s out of bounds",
				   SCINI_CONN_TIMEOUT);
	} else {
		SCACHE_G(ini_conn_timeout) = i;
	}
}

PHP_INI_MH(sc_ini_conn_timeout_changed) {
	sc_ini_conn_timeout_changed_i();
	return SUCCESS;
}


/* ==================================
 * byteorder helpers
 *  
 * */


/* network-to-host */
static inline int64_t
sc_ntohll(char *buffer) {
    return (((int64_t)buffer[0] << 56) & 0xff00000000000000LL) |
	(((int64_t)buffer[1] << 48) & 0xff000000000000LL) |
	(((int64_t)buffer[2] << 40) & 0xff0000000000LL) |
	(((int64_t)buffer[3] << 32) & 0xff00000000LL) |
	(((int64_t)buffer[4] << 24) & 0xff000000LL) |
	(((int64_t)buffer[5] << 16) & 0xff0000LL) |
	(((int64_t)buffer[6] << 8) & 0xff00LL) |
	((int64_t)buffer[7] & 0xffLL);
	
}

static inline int32_t
sc_ntohl(char *buffer) {
    return ((buffer[0] << 24) & 0xff000000) |
	    ((buffer[1] << 16) & 0xff0000) |
	    ((buffer[2] << 8) &0xff00) |
	    (buffer[3] & 0xff);
}

static inline int16_t
sc_ntohs(char *buffer) {
    
    return ((buffer[0] << 8) & 0xff00) |
		(buffer[1] & 0xff);
}

/* host-to-network */
static inline char *
sc_htonll(char *buffer, int64_t val) {

    buffer[7] = (char)(val & 0xff);
    buffer[6] = (char)((val >> 8) & 0xff);
    buffer[5] = (char)((val >> 16) & 0xff);
    buffer[4] = (char)((val >> 24) & 0xff);
    buffer[3] = (char)((val >> 32) & 0xff);
    buffer[2] = (char)((val >> 40) & 0xff);
    buffer[1] = (char)((val >> 48) & 0xff);
    buffer[0] = (char)((val >> 56) & 0xff);
    return buffer + 8;
}

static inline char *
sc_htonl(char *buffer, int32_t val) {

    buffer[3] = (char)(val & 0xff); 
    buffer[2] = (char)((val >> 8) & 0xff);
    buffer[1] = (char)((val >> 16) & 0xff);
    buffer[0] = (char)((val >> 24) & 0xff);
    return buffer + 4;
}

static inline char *
sc_htons(char *buffer, int16_t val) {

    buffer[1] = (char)(val & 0xff); 
    buffer[0] = (char)((val >> 8) & 0xff);
    return buffer + 2;
}


/* ====================================
 * sc_get_globalval
 *
 * fetch values from php's $GLOBALS
 *
 *   - global, like '_SERVER'
 *   - name,   like 'PHP_SELF'
 * */

static const char *
sc_get_globalval(char *global, char *name TSRMLS_DC) {

	zval **gzv;
	zval **kzv;

	if (zend_hash_find(&EG(symbol_table), global, strlen(global) + 1,
					   (void **)&gzv) == FAILURE) return NULL;

	if (Z_TYPE_PP(gzv) != IS_ARRAY) return NULL;
	if (zend_hash_find(Z_ARRVAL_PP(gzv), name, strlen(name) + 1,
					   (void **)&kzv) == FAILURE) return NULL;

	if (Z_TYPE_PP(kzv) != IS_STRING) return NULL;
	return Z_STRVAL_PP(kzv);
}

#if (0) /* for debugging */
static void
sc_dump_packet(int length, char *req) {

    int i;
    
    fprintf(stderr, "DUMP: %i", length);
    for(i = 0; i < length; i++) {
		if (!(i % 8)) printf("\n%p: ", req + i);
		fprintf(stderr, " %.2x", (unsigned int)(req[i] & 0xff));
    }
    fprintf(stderr, "\n");
}
#endif


static const char *
sc_strerror(enum scache_error_t scerr) {

	if ((scerr < 0) || (scerr >= sc_asize(scache_errstr))) {
		return "SCERR_UNKNOWN_ERROR";
	} else {
		return scache_errstr[scerr];
	}
}


/* ==========================================
 * connection startup, verification and close
 * ========================================== */
static int
sc_conn_connect_ip(const char *host, int port) {

	int rv, fdfl, sock;

#ifdef ENABLE_IPV6
	struct addrinfo hint, *ainfo;

	/* probably broken, never tested, tell me how this
	   actually should be done */

	memset(&hint, 0, sizeof(hint));
	hint.ai_family = AF_UNSPEC;
	hint.ai_socktype = 0;
	hint.ai_protocol = 0;
	hint.ai_flags = AI_ADDRCONFIG;

	if ((rv = getaddrinfo(host, NULL, &hint, &ainfo)) < 0) {
		if (rv == EAI_NONAME) {
			zend_error(E_WARNING, "host '%s' not found", host);
			return -1;
		} else {
			zend_error(E_WARNING,
					   "getaddrinfo() for host '%s' failed: %s",
					   host, gai_strerror(rv));
			return -1;
		}
	}

	if (((struct sockaddr_in *)(ainfo->ai_addr))->sin_family == AF_INET6) {
		((struct sockaddr_in6 *)(ainfo->ai_addr))->
			sin6_port = htons(port);
	} else {
		((struct sockaddr_in *)(ainfo->ai_addr))->
			sin_port = htons(port);
	}

	/* socket */
	sock = socket(((struct sockaddr_in *)(ainfo->ai_addr))->sin_family,
				  SOCK_STREAM, PF_UNSPEC);
	if (sock < 0) return -1;

#else
	struct hostent *he;
	struct sockaddr_in sa;

	/* resolve host */
	sa.sin_family = AF_INET;
	sa.sin_port = htons(port);
	sa.sin_addr.s_addr = inet_addr(host);

	if ((sa.sin_addr.s_addr == INADDR_NONE) &&
		(he = gethostbyname(host)) != NULL) {
		
		memcpy(&(sa.sin_addr.s_addr), he->h_addr_list[0], 
			   sizeof(sa.sin_addr.s_addr));
	} else {
		zend_error(E_ERROR, "can't resolve host: '%s'", host);
	}
	
	/* socket */
	sock = socket(AF_INET, SOCK_STREAM, PF_UNSPEC);
	if (sock < 0) return -1;
#endif
	
	/* mark temporarily nonblocking for connect */
	fdfl = fcntl(sock, F_GETFL);
	if (fdfl >= 0 && !fcntl(sock, F_SETFL, fdfl | O_NONBLOCK)) {

		do {
#ifdef ENABLE_IPV6
			rv = connect(sock, 
						 (struct sockaddr *)ainfo->ai_addr,
						 ainfo->ai_addrlen);
#else
			rv = connect(sock, (struct sockaddr *)&sa, sizeof(sa));
#endif

		} while(rv < 0 && errno == EINTR);

		if ((rv < 0) && (errno == EINPROGRESS)) {

			struct pollfd pfds[1];

#ifdef ENABLE_IPV6
			freeaddrinfo(ainfo);
#endif

			pfds[0].fd = sock;
			pfds[0].events = POLLOUT;

			do { /* poll until connected or timeout */
				rv = poll(pfds, 1, max_connect_timeout);
			} while(rv < 0 && errno == EINTR);
			
			if ((rv > 0) && (pfds[0].revents & POLLOUT)) {
				if (!fcntl(sock, F_SETFL, fdfl)) {

					int one = 1;
					if (setsockopt(sock, IPPROTO_TCP,
								   TCP_NODELAY, &one, sizeof(one)) < 0) {
						fprintf(stderr, "setsockopt: %s\n",
								strerror(errno));
					}

					return sock;
				}
			}
		}
	}
	
	while(close(sock) < 0 && errno == EINTR);
	zend_error(E_WARNING, "connection failed");
	
	return -1;
}

static int
sc_conn_connect_unix(const char *path) {

	int rv, sock;

	struct sockaddr_un sa;

	sa.sun_family = AF_UNIX;
	strncpy(sa.sun_path, path, sizeof(sa.sun_path));

	sock = socket(AF_UNIX, SOCK_STREAM, PF_UNSPEC);
	if (sock < 0) return -1;

	do {
		rv = connect(sock, (struct sockaddr *)&sa, SUN_LEN(&sa));
	} while(rv < 0 && errno == EINTR);

	if (rv < 0) {
		zend_error(E_WARNING, "connection failed: %s",
				   strerror(errno));
	}

	return sock;
}

static int
sc_conn_connect(const char *host, int port) {

	if (strchr(host, (int)'/')) {
		return sc_conn_connect_unix(host);
	} else {
		return sc_conn_connect_ip(host, port);
	}
}

static void
sc_conn_close(struct sc_conn_t *conn) {

	while(close(conn->fd) < 0 && errno == EINTR);
	conn->fd = -1;
	conn->dirty = 1;
}

struct sc_conn_t *
sc_conn_verify(struct sc_conn_t *conn) {

	if (conn->dirty) {
		sc_conn_close(conn);
		conn->fd = sc_conn_connect(conn->host, conn->port);
		conn->dirty = (conn->fd < 0) ? 1 : 0;
	}

	return conn;
}

static void
sc_conn_close_all(TSRMLS_C) {

	struct sc_conn_t *conn, *tmp;

	conn = SCACHE_G(sc_conn_list);
	while((tmp = conn)) {
		conn = conn->next;

		while(close(tmp->fd) < 0 && errno == EINTR);
		sc_pefree(tmp->host, 1);
		sc_pefree(tmp, 1);
	}
}

static struct sc_conn_t *
sc_conn_create(const char *host, int port,
			   int persistent TSRMLS_DC) {

	int sock;
	struct sc_conn_t **slot, *conn;

	/* default to port 0 if unix-sock */
	if (*host == '/') port = 0;

	/* if persistent requested, search from cache */
	if (persistent) {
		slot = &SCACHE_G(sc_conn_list);
		while((conn = *slot)) {
			
			if (conn->port == port &&
				!strcmp(conn->host, host)) {
				
				/* cache hit, move to first in queue */
				*slot = conn->next;
				conn->next = SCACHE_G(sc_conn_list);
				SCACHE_G(sc_conn_list) = conn;
			
				return sc_conn_verify(conn);
				
			} else {
				
				/* next */
				slot = &((*slot)->next);
			}
		}
	}

	/* not found, persistent or not */
	sock = sc_conn_connect(host, port);
	if (sock > 0) {

		conn = pecalloc(1, sizeof(struct sc_conn_t), 1);
		conn->host = sc_pestrdup(host, 1);

		if (persistent) {

			conn->persistent = 1;

			/* queue to cache */
			conn->next = SCACHE_G(sc_conn_list);
			SCACHE_G(sc_conn_list) = conn;
		}

		conn->fd = sock;
		conn->port = port;
		conn->dirty = 0;
		return conn;
	}

	return NULL;
}







/* ========================================
 * request handling, bottom layer, sock-io
 * ======================================== */
static enum scache_error_t
sc_post(struct sc_session_t *sess, 
		struct sc_wblock_t *wbfirst,
		int retry TSRMLS_DC) {

	int i, tlen, rv, count;
	struct iovec *iov;
	struct sc_wblock_t *wblock, *wbtemp, **wbptr;

	if (sc_conn_verify(sess->conn TSRMLS_CC) < 0) {

		zend_error(E_ERROR,
				   "cannot connect scache server.");
		return SCERR_NOT_CONNECTED; /* never */
	}

	/* mark as dirty, so when we break out in middle
	   it is left unstable to reconnect */
	sess->conn->dirty = 1;

	/* count blocks to write and total length */
	count = tlen = 0;
	wbptr = &wbfirst;
	while(*wbptr) {

		if ((*wbptr)->length == 0) {
			*wbptr = (*wbptr)->next;
			continue;
		}

		count++;
		tlen += (*wbptr)->length;
		(*wbptr)->pos = 0;
		wbptr = &((*wbptr)->next);
	}


	/* allocate iov, max_iov might be smaller
	   than total blocks to write */
	if (count > max_iov) count = max_iov;
	iov = alloca(count * sizeof(struct iovec));
	wbtemp = wbfirst;


	
	/* writeloop, loop until all written or error */
	while(wbtemp) { 


		wblock = wbtemp;

		/* collect iovec */
		for(i = 0; i < count && wblock; i++) {

			if (wblock->length < wblock->pos) { /* sanity check */

				zend_error(E_ERROR, 
						   "write position beyond buffer. internal "
						   "extension error, not yours");

				return SCERR_SYSTEM; /* never */
			}
			iov[i].iov_base = wblock->base + wblock->pos;
			iov[i].iov_len = wblock->length - wblock->pos;
			wblock = wblock->next;
		}


		do { /* EINTR */
			rv = writev(sess->conn->fd, iov, i);
		} while(rv < 0 && errno == EINTR);


		if (rv > 0) { 

			/* ff to next unwritten position, in case
			   of count of wblocks being greater than
			   max_iov, complete write needs several writev()s */
			while(rv > 0) {

				wblock = wbtemp;
				
				/* assert */
				if (!wblock) {
					zend_error(E_ERROR, 
							   "null block in write-queue. internal "
							   "extension error, not yours");

					return SCERR_SYSTEM;
				}

				i = wblock->length - wblock->pos;
				if (rv >= i) {
					rv -= i;
					wbtemp = wbtemp->next;
				} else {
					wblock->pos += rv;
					continue;
				}
			}

		} else if (retry > 0) {

			/* recurse and retry again. as side effect
			   conn->dirty is set and we reconnect.

			   server hasn't operated anything because it
			   expects whole request before continuing */
			return sc_post(sess, wbfirst, retry - 1 TSRMLS_CC);

		} else {

			/* too many retries */
			zend_error(E_ERROR, 
					   "request failed, too "
					   "many retries.");
			return SCERR_PROTOCOL; /* never */
		}
	}

	sess->conn->dirty = 0; /* sock is valid again */

	return SCERR_SUCCESS;
}

static int
sc_read_buffer(int fd, char *buffer, int buflen) {

	int rv, rb = 0;
	struct pollfd pfd[1];

	pfd[0].fd = fd;
	pfd[0].events = POLLIN;

	while(buflen > 0) {

		/* try read */
		do {
			rv = read(fd, buffer, buflen);
		} while((rv < 0) && (errno == EINTR));

		if (rv > 0) { 

			buflen -= rv;
			buffer += rv;
			rb += rv;

			if (buflen > 0) {

				/* poll for more */
				rv = poll(pfd, 1, max_connect_timeout);
				if (rv > 0) continue;

				zend_error(E_ERROR, 
						   "connection timed out when waiting "
						   "server response.");

				/* dies */
			}

		} else if (rv == 0) {

			/* EOF */
			zend_error(E_ERROR,
					   "server has closed connection. request "
					   "failed.");

			return -1; /* never */

		} else {

			zend_error(E_ERROR,
					   "reading server response failed: %s",
					   strerror(errno));
			return -1; /* never */
		}
	}

	return rb;
}

static enum scache_error_t
sc_read(struct sc_session_t *sess,
		struct sc_req_t *req,
		int req_cnt TSRMLS_DC) {

	int rv, res_count, res_len, res_flags, res_index;
	char hdr[12];

	sess->conn->dirty = 1;

	/* read header */
	rv = sc_read_buffer(sess->conn->fd, hdr, sizeof(hdr));
	if (rv == sizeof(hdr)) {

		/* parse protocol header */
		res_count = res_len = sc_ntohl(hdr);
		res_index = sc_ntohl(hdr + 4);
		res_flags = sc_ntohl(hdr + 8);

		res_count = (res_count >> 24) & 0x7f;
		res_len = res_len & 0xffffff;
		
		res_flags  &= SCACHE_RFLAG_MASK;
		req->flags &= SCACHE_QFLAG_MASK;
		req->flags |= res_flags;

		sess->index = res_index; /* session index is always collected */

		/* replace buffer if required */
		res_len -= sizeof(hdr);
		if (!req->rbuffer || (req->rlen < res_len)) {
			req->rbuffer = sc_emalloc(res_len + _NMTOBO);
		}

		/* read body */
		req->rlen = res_len;
		rv = sc_read_buffer(sess->conn->fd, req->rbuffer, req->rlen);

		if ((rv = req->rlen) && (res_count == req_cnt)) {

			sess->conn->dirty = 0;
			sess->scerr = SCERR_SUCCESS;
			return sess->scerr;
		}

	}

	/* force close and fail miserably */
	sc_conn_close(sess->conn);
	sess->scerr = SCERR_PROTOCOL;

	zend_error(E_ERROR,
			   "reading response failed, packet length mismatch");

	return sess->scerr; /* never, zend_error panics */
}









/* ========================================
 * request handling, middle layer,
 * request-preparation 
 * ======================================== */

static enum scache_error_t 
sc_query(struct sc_session_t *sess,
		 struct sc_req_t *req TSRMLS_DC) {

	int tlen = 0;
	char *buffer;
	struct sc_wblock_t *whead, *wreq;

	/* protocol header, calculate total length */
	for(wreq = req->wblocks; wreq; wreq = wreq->next) {
		tlen += wreq->length;
	}

	whead = sess->header; /* already created in sc_session_alloc() */
	whead->next = wreq = alloca(sizeof(struct sc_wblock_t));

	/* req-type -header */
	tlen += wreq->length = 4;
	wreq->base = alloca(wreq->length);
	wreq->next = req->wblocks;
	sc_htonl(wreq->base, SC_HDR_PACK_R(req->reqtype, tlen));

	/* update stored protocol header */

	/* DEPENDS: sc_iov_exec (PROTOHEADER) 
	 * 
	 * sc_iov_exec constructs main header in own, so if you
	 * decide to change protocol, change also sc_iov_exec
	 * (and backend) */

	tlen += whead->length;
	buffer = whead->base;
	buffer = sc_htonl(buffer, SC_HDR_PACK_H(1, tlen));	
	buffer = sc_htonl(buffer, sess->index);
	buffer = sc_htonl(buffer, req->flags);

	sess->scerr = sc_post(sess, whead, max_retries TSRMLS_CC);
	if (sess->scerr == SCERR_SUCCESS) {

		sess->scerr = sc_read(sess, req, 1 TSRMLS_CC);
		/* actually other than SCERR_SUCCESS doesn't return, but
		   zend_errors */
	}

	/* so there is only success */
	return sess->scerr;
}

static enum scache_error_t
sc_parse_simple_response(struct sc_req_t *req, 
						 int *rlen, char **rbuf) {

	int restype;
	int reslength;
	int reserr;

	*rlen = 0;
	*rbuf = NULL;
	
	if (!req->rbuffer || req->rlen < 4) {

		zend_error(E_ERROR, "server sent short reply");
		return SCERR_PROTOCOL; /* never */
	}
	
	restype = reslength = sc_ntohl(req->rbuffer);
	restype = (restype >> 24) & 0xff;
	reslength = reslength & 0xffffff;

	if (req->rlen != reslength) {
		zend_error(E_ERROR, "server reply does not match expected size");
		return SCERR_PROTOCOL; /* never */
	}
	
	if (req->reqtype != (restype & 0x7f)) {
		zend_error(E_ERROR, "server response does not match request sent");
		return SCERR_PROTOCOL; /* never */
	}

	if (restype & 0x80) {

		if (req->rlen < 8) {
			zend_error(E_ERROR, "server sent error with short packet");
			return SCERR_PROTOCOL; /* never */
		}

		reserr = sc_ntohl(req->rbuffer + 4);
		*rlen = req->rlen - 8;
		*rbuf = req->rbuffer + 8;

		return reserr;

	} else {

		*rlen = req->rlen - 4;
		*rbuf = req->rbuffer + 4;

		return SCERR_SUCCESS;
	}
}





/* ========================================
 * session create, manage and destroy
 * ======================================== */
static void
sc_session_clear(struct sc_session_t *sess) {

	if (sess->conn) {
		if (!sess->conn->persistent) {

			sc_conn_close(sess->conn);
			sc_pefree(sess->conn->host, 1);
			sc_pefree(sess->conn, 1);
			sess->conn->host = NULL;
			sess->conn = NULL;
		}
	}

	if (sess->header) {
		sc_pefree(sess->header->base, 1);
		sc_pefree(sess->header, 1);
		sc_pefree(sess->secret, 1);
		sc_pefree(sess->tag, 1);
		sc_pefree(sess, 1);
	}
}

static void
sc_session_dtor(zend_rsrc_list_entry *rsrc TSRMLS_DC) {
	return sc_session_clear((struct sc_session_t *)rsrc->ptr);
}

/* resolve automaticly inserted suffix for session tag */
static char *
sc_remap_tag(const char *tag) {

	int tlen;
	char *cp, *ident;

	tlen = strlen(tag);
	if (!(ident = SCACHE_G(ini_tag_suffix))) {

		return sc_pestrdup(tag, 1);
	}

	if (!strncmp(ident, "auto:", 5)) {

		struct stat st;
		const char *self;
		char buffer[64];

		ident += 5;
		*buffer = 0;

		if (!strcmp(ident, "exec-euid")) {
			sprintf(buffer, "__EUID:%u+__", geteuid());
		} else if (!strcmp(ident, "exec-egid")) {
			sprintf(buffer, "__EGID:+%u__", getegid());
		} else if (!strcmp(ident, "exec-eugid")) {
			sprintf(buffer, "__EGID:%u+%u__", geteuid(), getegid());
		} else if ((self = sc_get_globalval("_SERVER", "PHP_SELF")) &&
				   (!VCWD_STAT(self, &st))) {

			if (!strcasecmp(ident, "uid")) {
				sprintf(buffer, "__%u+__", st.st_uid);
			} else if (!strcasecmp(ident, "gid")) {
				sprintf(buffer, "__+%u__", st.st_gid);
			} else if (!strcasecmp(ident, "ugid")) {
				sprintf(buffer, "__%u+%u__", st.st_uid, st.st_gid);
			}
		}

		if (!*buffer) strcpy(buffer, "__ERROR__");
		tlen += strlen(buffer) + 2 + _NMTOBO;

		sprintf((cp = sc_pemalloc(tlen, 1)), "%s:%s", tag, buffer);
		return cp;

	} else if ((cp = strchr(ident, (int)':'))) {

		/* static: -suffix */
		tlen += strlen(cp) + 2 + _NMTOBO;

		/* ...to append */
		sprintf((cp = sc_pemalloc(tlen, 1)), "%s:%s", 
				 tag, strchr(ident, (int)':') + 1);
		return cp;

	} else {

		tlen += strlen(ident) + 2 + _NMTOBO;
		sprintf((cp = sc_pemalloc(tlen, 1)), "%s:%s", 
				 tag, ident);
		return cp;
	}
}

static struct sc_session_t *
sc_session_alloc(struct sc_conn_t *conn, 
				 const char *tag, const char *secret) {

	int taglen, seclen, athlen;
	char *auth, *buffer;
	struct sc_session_t *sess;

	/* auth-string */
	auth = SCACHE_G(ini_auth_string);
	if (!auth) auth = "";

	sess = sc_pecalloc(1, sizeof(struct sc_session_t), 1);
	sess->index = -1;
	sess->conn = conn;
	sess->tag = sc_remap_tag(tag); //pestrdup(tag, 1);
	sess->secret = sc_pestrdup(secret, 1);

	taglen = strlen(sess->tag) + 1;
	seclen = strlen(sess->secret) + 1;
	athlen = strlen(auth) + 1;

	/* is length of result of sc_remap_tag valig */
	if (strlen(tag) < SCACHE_PTO_MIN_TAGLEN) {
		zend_error(E_WARNING, 
				   "after suffixing session identifier is too short");
	} else if (strlen(tag) > SCACHE_PTO_MAX_TAGLEN) {
		zend_error(E_WARNING, 
				   "after suffixing session identifier is too long");
	}
	
	/* precreate protocol header template on
	   using data */
	sess->header = sc_pemalloc(sizeof(struct sc_wblock_t), 1);
	sess->header->length = 18 + taglen + seclen + athlen;
	sess->header->base = buffer = 
		sc_pemalloc(sess->header->length + _NMTOBO, 1);
	
	/* prepare parts that are static */
	buffer = sc_htons(buffer + 12, taglen);
	buffer = sc_htons(buffer, seclen);
	buffer = sc_htons(buffer, athlen);

	memcpy(buffer, sess->tag, taglen);
	buffer += taglen;

	memcpy(buffer, sess->secret, seclen);
	buffer += seclen;

	memcpy(buffer, auth, athlen);
	buffer += seclen;

	return sess;
}


static struct sc_session_t *
sc_connect_real(int idx, const char *host, const char *tag,
				const char *secret, int port, 
				int persistent TSRMLS_DC) {

	struct sc_conn_t *conn;

	/* prepare null parameters */
	if (!host || !*host) host = SCACHE_G(ini_default_host);
	if (!host || !*host) host = SCACHE_PTO_UNIX_SOCK;

	if (!port) port = SCACHE_G(ini_default_port);
	if (!port) port = SCACHE_PTO_LISTEN_PORT;
	if (!secret) secret = "";


	/* verify parameters */
	if (!host || !*host) {
		zend_error(E_WARNING, "host missing or empty");
	} else if (!tag || !*tag) {
		zend_error(E_WARNING, "session identifier missing or empty");
	} else if (!secret) {
		zend_error(E_WARNING, "invalid secret");
	} else if (strlen(tag) < SCACHE_PTO_MIN_TAGLEN) {
		zend_error(E_WARNING, "session identifier is too short");
	} else if (strlen(tag) > SCACHE_PTO_MAX_TAGLEN) {
		zend_error(E_WARNING, "session identifier is too long");
	} else if (strlen(secret) < SCACHE_PTO_MIN_SECRETLEN) {
		zend_error(E_WARNING, "session secret is too short");
	} else if (strlen(secret) > SCACHE_PTO_MAX_SECRETLEN) {
		zend_error(E_WARNING, "session secret is too long");
	} else if (port < SCACHE_PTO_MIN_PORT || port > SCACHE_PTO_MAX_PORT) {
		zend_error(E_WARNING, "invalid port %i", port);
	} else {
		goto param_ok;
	}

	return NULL;

 param_ok:

	conn = sc_conn_create(host, port, persistent TSRMLS_CC);
	if (conn) return sc_session_alloc(conn, tag, secret);

	return NULL;
}

static void
sc_connect_create_res(struct sc_session_t *sess,
					  zval *return_value TSRMLS_DC) {

	/* register session and return. 
	 *
	 * no sock-io was done during whole operation, just
	 * connection was created and session-information stored
	 * for actual future request
	 *
	 * in case that session does not exist, caller is
	 * notified on first operation */

	if (sess) {

		int rid;
		zval *zv;

		MAKE_STD_ZVAL(zv);
		rid = ZEND_REGISTER_RESOURCE(zv, (void *)sess, le_sessions);
		RETURN_RESOURCE(rid);

	} else {

		/* never */
		zend_error(E_WARNING, "cannot register resource, no session");
		RETURN_FALSE;
	}
}


/* ===================================================
 * sc_connect
 *
 *  - create session-connection resource
 *  - prepare queryheader
 *
 * sc_connect -family is a fast connnection method,
 * that does not query server, but just prepares
 * client for following queries
 *
 * session need to be created by sc_open
 *
 */

static void 
sc_connect(INTERNAL_FUNCTION_PARAMETERS, int persistent) {

	int host_len, tag_len, sec_len;
	long port;
	char *host, *tag, *secret;

	/* setup defaults */
	port = 0;
	host = tag = secret = NULL;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
							  "s|ssl", 
							  &tag, &tag_len,
							  &host, &host_len,
							  &secret, &sec_len,
							  &port) == FAILURE) RETURN_FALSE;

	struct sc_session_t *sess =
		sc_connect_real(-1, host, tag, secret, port,
						persistent TSRMLS_CC);

	if (sess) {
		sc_connect_create_res(sess, return_value TSRMLS_CC);
	} else {
		RETURN_FALSE;
	}
}

/* connect_fast -functions are actually equal to plain connect, it
 * just have slot for specifying server side index.
 *
 * on server, the idx is only first to lookup, it server doesn't find
 * given session that given index-slot, it searches it from btree
 * anyway, the thing that scache_connect() does.
 *
 * so, scache_connect('session') is just as fast as
 * scache_connect('session', random-index). it is fast version
 * when given index really belongs to right session 
 *
 * you can freely use sc_connect_fast and submit invalid index
 * to it as long as you don't know the right one
 *
 */

static void
sc_connect_fast(INTERNAL_FUNCTION_PARAMETERS, int persistent) {


	int host_len, tag_len, sec_len;
	long idx, port;
	char *host, *tag, *secret;

	/* setup defaults */
	port = 0;
	host = tag = secret = NULL;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
							  "sl|ssl", 
							  &tag, &tag_len,
							  &idx,
							  &host, &host_len,
							  &secret, &sec_len,
							  &port) == FAILURE) RETURN_FALSE;


	sc_connect_create_res(sc_connect_real(idx, host, tag, secret, port, 
										  persistent TSRMLS_CC),
						  return_value TSRMLS_CC);
}

static enum scache_error_t
sc_open_session(struct sc_session_t *sess, 
				int flags, int expire TSRMLS_DC) {
	
	char *buffer, *stackbuf;
	struct sc_req_t req;
	struct sc_wblock_t wblock;

	/* prepare req-content */
	wblock.length = 8;
	wblock.base = buffer = alloca(wblock.length + _NMTOBO);
	wblock.next = NULL;

	buffer = sc_htonl(buffer, expire);
	buffer = sc_htonl(buffer, flags);

	/* prepare req-hdr */
	req.reqtype = SCREQ_TYPE_SESSION_CREATE;
	req.flags = 0;
	req.rlen = 256; /* expected, gets replaced if insufficient */
	req.rbuffer = stackbuf = alloca(req.rlen);
	req.wblocks = &wblock;

	/* do request */
	if (sc_query(sess, &req TSRMLS_CC) == SCERR_SUCCESS) {

		int rlen;
		char *rbuf;

		sess->scerr = sc_parse_simple_response(&req, &rlen, &rbuf);
	}

	if (req.rbuffer != stackbuf) {

		/* got replaced, we were trying to operate readwrite on
		   stack, but buffer we allocated was not large enough
		   and got replaces with emalloced larger one on
		   sc_read */
		sc_efree(req.rbuffer);
	}

	return sess->scerr;
}


/* ===========================================================
 * sc_open
 *
 *   - open and verify connection
 * 
 * sc_open should actually be considered as sc_create. it does a round
 * trip on server, and creates session with tag and secret
 * specified. but having non-overwrite-flag on header id, it does not
 * overwrite server session but merely bounces back, so as a side
 * effect sc_open verifies existence of session specified.
 *
 * sc_connect-family is just fast local operation that stores
 * connection information for further requests.
 *
 * right and fastest way to do things is create session with
 * sc_reset() or sc_open() when user logs in, pick its index and store
 * it to cookie or other session tracker and sc_pconnect_fast() in
 * further page accesses. ie. get index, make cookie like
 * "index/the-session-tag", split it back on further request for
 * connecting with sc_pconnect_fast(index/session-tag)
 * 
 */

static void
sc_open(INTERNAL_FUNCTION_PARAMETERS, int persistent, int overwrite) {

	int host_len, tag_len, sec_len;
	long port, flags, expire;
	char *tag, *host, *secret;

	/* setup defaults */
	port = expire = flags = 0;
	host = tag = secret = NULL;

	struct sc_session_t *sess;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
							  "s|sslll", 
							  &tag, &tag_len,
							  &host, &host_len,
							  &secret, &sec_len,
							  &expire,
							  &port,
							  &flags) == FAILURE) RETURN_FALSE;


	/* create null connection */
	sess = sc_connect_real(-1, host, tag, secret, port, 
						   persistent TSRMLS_CC);

	if (sess) {

		/* verify sensible defaults */
		if (expire == 0) expire = SCACHE_PTO_DEF_EXPIRE;
		if (expire < SCACHE_PTO_MIN_EXPIRE) {
			zend_error(E_WARNING, 
					   "Given expire time %li seconds is below protocol "
					   "minimum. Using protocol minumum %i seconds",
					   expire, SCACHE_PTO_MIN_EXPIRE);
			expire = SCACHE_PTO_MIN_EXPIRE;
		}
		if (expire > SCACHE_PTO_MAX_EXPIRE) {
			zend_error(E_WARNING, 
					   "Given expire time %li seconds is over protocol "
					   "maximum. Using protocol maximum %i seconds",
					   expire, SCACHE_PTO_MAX_EXPIRE);
			expire = SCACHE_PTO_MAX_EXPIRE;
		}


		if (overwrite) {
			flags |= SCACHE_QFLAG_OVERWRITE;
		} else {
			flags &= ~SCACHE_QFLAG_OVERWRITE;
		}

		/* exec */
		switch(sc_open_session(sess, flags, expire TSRMLS_CC)) {
		case SCERR_SUCCESS:
			
			/* ok, store as resource */
			sc_connect_create_res(sess, return_value TSRMLS_CC);
			return;

		case SCERR_EXIST:
			zend_error(E_WARNING, 
					   "session already exist and can't "
					   "be overwritten!");
			break;

		case SCERR_LIMITS_REACHED:
			zend_error(E_WARNING,
					   "per user or per host limits reached!");
			break;

		case SCERR_PROTOCOL:
			zend_error(E_WARNING, "invalid request, protocol error!");
			break;

		default:
			zend_error(E_WARNING, "server failed!");
			break;
		}

	} else {
		/* zend_error notified already */
	}

	RETURN_FALSE;
}

static void
sc_open_m(INTERNAL_FUNCTION_PARAMETERS, int overwrite) {

	long flags = 0;
	struct sc_classdata_t *cdata = (struct sc_classdata_t *)
		zend_object_store_get_object(getThis() TSRMLS_CC);

	if (overwrite) {
		flags |= SCACHE_QFLAG_OVERWRITE;
	} else {
		flags &= ~SCACHE_QFLAG_OVERWRITE;
	}

	switch(sc_open_session(cdata->session, flags,
						   cdata->expire TSRMLS_CC)) {
	case SCERR_SUCCESS:
		/* ok */
		RETURN_TRUE;
		return;
	case SCERR_EXIST:
		zend_error(E_WARNING, 
				   "session already exist and can't "
				   "be overwritten!");
		break;
	case SCERR_LIMITS_REACHED:
		zend_error(E_WARNING,
				   "per user or per host limits reached!");
		break;
	case SCERR_PROTOCOL:
		zend_error(E_WARNING, "invalid request, protocol error!");
		break;
	default:
		zend_error(E_WARNING, "server failed!");
		break;
	}

	RETURN_FALSE;
}



/* ========================================
 * request handling, top layer,
 * ======================================== */


static int
sc_verify_arg_r(const char *funcname, zval *zv, const char *argid) {

	int len;

	if (Z_TYPE_P(zv) == IS_LONG) convert_to_string(zv);
	if ((Z_TYPE_P(zv) != IS_STRING) || (Z_STRVAL_P(zv) == NULL)) {

		zend_error(E_ERROR,
				   "%s expects parameter %s (key) to be non-null "
				   "string.", funcname, argid);
		return -1;
	}

	len = strlen(Z_STRVAL_P(zv));
	if (len < 1 ) {
		/* empty key is reserved from session_handler */
		zend_error(E_ERROR,
				   "%s: key specified is empty", funcname);

	} else if (len > SCACHE_PTO_MAX_KEYLEN) { /* too long */

		zend_error(E_ERROR,
				   "%s: key specified is too long. "
				   "max length is %u, keylength was %u.",
				   funcname, SCACHE_PTO_MAX_KEYLEN, len);

	} else if (len != Z_STRLEN_P(zv)) {

		zend_error(E_ERROR,
				   "%s: key has nul bytes in middle. key needs to "
				   "be plain nul-terminated string, not "
				   "binary string.", funcname);
	} else {

		/* ok */
		return 0;
	}
		
	/* fallback fail */
	return -1;
}

static int
sc_verify_as_key2(const char *funcname, zval *zv) {
	return sc_verify_arg_r(funcname, zv, "2");
}

static int
sc_verify_as_key3(const char *funcname, zval *zv) {
	return sc_verify_arg_r(funcname, zv, "3");
}


static int
sc_serialize_arg3(char *funcname, zval *zv, struct sc_iop_t *iop) {

	php_serialize_data_t sd;
	smart_str buf = {0};

	/* remember to use binary-safe length and include
	   nul termination in total length */

	if (Z_TYPE_P(zv) == IS_STRING) {

		iop->arg2_prefix = 'b'; /* binary */
		iop->arg2 = Z_STRVAL_P(zv);
		iop->arg2_len = Z_STRLEN_P(zv);

		if (iop->arg2_len > SCACHE_PTO_MAX_VALUELEN) {
			zend_error(E_ERROR,
					   "%s: max content size (%i) exceeded", 
					   funcname, SCACHE_PTO_MAX_VALUELEN);
			return -1;
		}

		return 0;

	} else {

		PHP_VAR_SERIALIZE_INIT(sd);
		php_var_serialize(&buf, &zv, &sd TSRMLS_CC);
		PHP_VAR_SERIALIZE_DESTROY(sd);
		
		if (buf.c) {
			
			iop->arg2_prefix = 's'; /* serialized */
			iop->arg2 = buf.c;
			iop->arg2_len = buf.len; 

			if (iop->arg2_len > SCACHE_PTO_MAX_VALUELEN) {
				zend_error(E_ERROR,
						   "%s: max content size (%i) exceeded", 
						   funcname, SCACHE_PTO_MAX_VALUELEN);
				return -1;
			}

			return 0;
		}
	}

	zend_error(E_ERROR,
			   "%s: serializing argument 3 failed", funcname);
	return -1; /* never */
}
				  


/* ========================================
 * sc_simple_command()
 *
 *  - run one iop 
 * */

static enum scache_error_t 
sc_simple_command(INTERNAL_FUNCTION_PARAMETERS,
				  char *funcname, struct sc_req_t *req) {


	zval *zv;
	struct sc_session_t *sess;

	if (!getThis()) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
								  "r", &zv) == FAILURE) {
			return SCERR_INVALID_REQUEST;
		}
		sess = (struct sc_session_t *)
			zend_fetch_resource(&zv , -1, SC_SESS_RESNAME, NULL, 1,
								le_sessions);
	} else { /* class-interface */

		struct sc_classdata_t *cdata = (struct sc_classdata_t *)
			zend_object_store_get_object(getThis() TSRMLS_CC);
		sess = cdata->session;
	}

	if (sess) {
		
		int nulval;
		char *nulbuf;

		req->wblocks = NULL;
		if (sc_query(sess, req TSRMLS_CC) == SCERR_SUCCESS) {
			sess->scerr = sc_parse_simple_response(req, &nulval, &nulbuf);
		}

	} else {
		zend_error(E_WARNING, "%s: no session", funcname);
	}

	return sess->scerr;
}



/* =========================================
 * iop functions for sending multiple
 * requests in one packet 
 *
 * */


/* =========================================
 * sc_iov_extract_iop
 *
 * - extract iops from php-array
 *
 * */
static struct sc_iop_t *
sc_iov_extract_iop(HashTable *ahash, HashPosition *pos, 
				   struct sc_iop_t *iop) {

	zval **zv1, **zv2, **zv3;

	/* fetch iop */
	if (zend_hash_get_current_data_ex(ahash, 
									  (void **)&zv1, pos) != SUCCESS) {
		goto fail_arg;
	} else if (Z_TYPE_PP(zv1) == IS_LONG ||
			   Z_TYPE_PP(zv1) == IS_STRING) {

		/* expect long or something convertable to */
		if (Z_TYPE_PP(zv1) == IS_STRING ||
			Z_TYPE_PP(zv1) == IS_DOUBLE) convert_to_long(*zv1);
		if (Z_TYPE_PP(zv1) != IS_LONG) goto fail_arg;
		iop->iop = Z_LVAL_PP(zv1);

	} else {
		goto fail_arg;
	}


	/* fetch key path or token, all but SCACHE_IOP_LASTERR
	   has key */
	if (iop->iop != SCACHE_IOP_LASTERR) {
		zend_hash_move_forward_ex(ahash, pos);
		if (zend_hash_get_current_data_ex(ahash,
										  (void **)&zv2, pos) == SUCCESS) {
			
			if (sc_verify_as_key2("sc_iov", *zv2) < 0) goto fail_arg;
			iop->arg1 = Z_STRVAL_PP(zv2);
			
		} else {
			goto fail_arg;
		}
	} else {

		/* no-arg */
		iop->arg1 = ""; /* dummy */
	}

	/* fetch arg2 depending of iop*/
	switch(iop->iop) {
	case SCACHE_IOP_SET:
	case SCACHE_IOP_ADD:
	case SCACHE_IOP_REPLACE:

	case SCACHE_IOP_SHSET:
	case SCACHE_IOP_SHADD:
	case SCACHE_IOP_SHREPLACE:

	case SCACHE_IOP_CHPUT:

	case SCACHE_IOP_RNPUSH:
	case SCACHE_IOP_RNUNSHIFT:
		
		/* expects string arg, so we serialize if
		   arg is not string. 

		   both are prepended with char to instruct
		   fetch functions to unserialize if original
		   was not string */

		zend_hash_move_forward_ex(ahash, pos);
		if (zend_hash_get_current_data_ex(ahash,
										  (void **)&zv3, pos) == SUCCESS) {

			if (!sc_serialize_arg3("scache_iov", *zv3, iop)) break;

		} else {

			/* accumulator */
			iop->arg2_prefix = 0;
			iop->arg2_len = -1; //0; /* protocol-null */
			iop->arg2 = "";
			break;
		}

		goto fail_arg;
		break;

	case SCACHE_IOP_COLL:
	case SCACHE_IOP_SHCOLL:

		zend_hash_move_forward_ex(ahash, pos);
		if (zend_hash_get_current_data_ex(ahash,
										  (void **)&zv3, pos) == SUCCESS) {

			if (sc_verify_as_key3("sc_iov", *zv3) < 0) goto fail_arg;
			iop->arg2 = Z_STRVAL_PP(zv3);
			iop->arg2_len = Z_STRLEN_PP(zv3) + 1;
			iop->arg2_prefix = 0;
			break;
		}

		goto fail_arg;
		break;
		
	case SCACHE_IOP_UNSET:
	case SCACHE_IOP_GET:
	case SCACHE_IOP_STAT:

	case SCACHE_IOP_SHUNSET:
	case SCACHE_IOP_SHGET:
	case SCACHE_IOP_SHSTAT:

	case SCACHE_IOP_CHGET:
	case SCACHE_IOP_CHCLEAR:

	case SCACHE_IOP_VGET:
	case SCACHE_IOP_VUNSET:

	case SCACHE_IOP_RNGET:
	case SCACHE_IOP_RNPOP:
	case SCACHE_IOP_RNSHIFT:
	case SCACHE_IOP_RNCLEAR:
	case SCACHE_IOP_RNSIZE:

	case SCACHE_IOP_LASTERR:

		/* iops above does not have second argument. 2nd argument
		   anyway is always present in packet, but empty */
		iop->arg2 = "";
		iop->arg2_prefix = 0;
		iop->arg2_len = 0;
		break;

	case SCACHE_IOP_VADD:
	case SCACHE_IOP_VSUB:
	case SCACHE_IOP_VSET:
	case SCACHE_IOP_VBRK:

	case SCACHE_IOP_RNROTF:
	case SCACHE_IOP_RNROTB:

		/* iops above have integer argument, so we expect
		   lval or something that can be converted to */

		zend_hash_move_forward_ex(ahash, pos);
		if (zend_hash_get_current_data_ex(ahash,
										  (void **)&zv3, pos) != SUCCESS) {

			goto fail_arg;
		} 
		
		/* expect long or something convertable to */
		if ((Z_TYPE_PP(zv3) == IS_STRING) ||
			(Z_TYPE_PP(zv3) == IS_DOUBLE) ||
			(iop->iop == SCACHE_IOP_VBRK)) convert_to_long(*zv3);

		if (Z_TYPE_PP(zv3) != IS_LONG) goto fail_arg;

		/* pack as 64bit nbo (as usual) */
		iop->arg2 = sc_emalloc(8);
		iop->arg2_prefix = 0;
		iop->arg2_len = 8;
		sc_htonll(iop->arg2, (int64_t)Z_LVAL_PP(zv3));
		break;

	default:

		/* unknown iop */
		zend_error(E_ERROR, "invalid operation: %lu", iop->iop);
		return NULL; /* never */
	}

	iop->next = NULL;
	return iop;

 fail_arg:

	return NULL; /* never */
}


/* =====================================================
 * sc_iov_result
 *
 *  - parse responses to bulk of iop requests
 * 
 * */

static int
sc_iov_result(struct sc_session_t *sess,
			  struct sc_req_t *req,
			  struct sc_iop_t *iops TSRMLS_DC) {

	int rlen;
	char *rbuf;

	struct sc_iop_t *iop;

	rlen = req->rlen;
	rbuf = req->rbuffer;
	if (!rbuf) {

		/* why we test that? */
		zend_error(E_ERROR,
				   "should not happen, no response buffer. "
				   "extension error, not yours.");
	}

	/* walk trough iops */
	for(iop = iops; iop; iop = iop->next) {

		int restype, reslen;

		iop->rval = NULL;

		/* response-header */
		if (rlen < 4) {
			zend_error(E_ERROR, 
					   "iov_result: no more data, response "
					   "packet (%i) misaligned\n", req->rlen);
			return -1; /* never */
		}

		restype = reslen = sc_ntohl(rbuf);
		restype = (restype >> 24) & 0xff;
		reslen = reslen & 0xffffff;

		if (iop->iop != (restype & 0x7f)) {

			zend_error(E_ERROR,
					   "iov_result: response received does not "
					   "match request: %lu requested, got %u",
					   iop->iop, (restype & 0x7f));
			return -1; /* never */

		} else if (restype & 0x80) { /* got error */

			/* request resulted error (which might be just fine
			   for for boolean operators, where error code might
			   be SCERR_SUCCEESS */
			if (rlen < 8) {
				zend_error(E_ERROR,
						   "iov_result: server sent short "
						   "error packet");
				return -1; /* never */
			}


			/* get error from packet */
			sess->scerr = iop->rerr = sc_ntohl(rbuf + 4);

			/* and resolve */
			switch((restype & 0x7f)) {
			case SCACHE_IOP_ADD:
			case SCACHE_IOP_SET:
			case SCACHE_IOP_REPLACE:
			case SCACHE_IOP_COLL:

			case SCACHE_IOP_SHADD:
			case SCACHE_IOP_SHSET:
			case SCACHE_IOP_SHREPLACE:
			case SCACHE_IOP_SHCOLL:

			case SCACHE_IOP_CHPUT:

			case SCACHE_IOP_VADD:
			case SCACHE_IOP_VSUB:
			case SCACHE_IOP_VSET:

			case SCACHE_IOP_RNPUSH:
			case SCACHE_IOP_RNUNSHIFT:
			case SCACHE_IOP_RNSIZE:
			case SCACHE_IOP_RNROTF:
			case SCACHE_IOP_RNROTB:
			case SCACHE_IOP_RNGET:
			case SCACHE_IOP_RNPOP:
			case SCACHE_IOP_RNSHIFT:
			case SCACHE_IOP_LASTERR: /* never */

				if (iop->rerr == SCERR_NO_ACCESS) {
					zend_error(E_WARNING,
							   "access not allowed. "
							   "host is not permitted to request "
							   "specified.");
				}

				/* boolean rval, success or not */
				MAKE_STD_ZVAL(iop->rval);
				ZVAL_BOOL(iop->rval,
						  (iop->rerr == SCERR_SUCCESS) ? 1 : 0);
				break;

			case SCACHE_IOP_UNSET:
			case SCACHE_IOP_SHUNSET:
			case SCACHE_IOP_CHCLEAR:
			case SCACHE_IOP_VUNSET:
			case SCACHE_IOP_RNCLEAR:

				if (iop->rerr == SCERR_NO_ACCESS) {
					zend_error(E_WARNING,
							   "access not allowed. "
							   "host is not permitted to request "
							   "specified.");
				}

				/* boolean rval, success or not */
				MAKE_STD_ZVAL(iop->rval);
				ZVAL_BOOL(iop->rval,
						  ((iop->rerr == SCERR_SUCCESS) ||
						   (iop->rerr == SCERR_NOTEXIST)) ? 1 : 0);
				break;

			case SCACHE_IOP_GET: 
			case SCACHE_IOP_SHGET:
			case SCACHE_IOP_CHGET:
			case SCACHE_IOP_VGET:

				if (iop->rerr == SCERR_NO_ACCESS) {
					zend_error(E_WARNING,
							   "access not allowed. "
							   "host is not permitted to request "
							   "specified.");
				}

				/* not found, that's ok, results false */
				MAKE_STD_ZVAL(iop->rval);
				ZVAL_BOOL(iop->rval, 0);
				break;

			case SCACHE_IOP_VBRK:

				/* true if broke */
				MAKE_STD_ZVAL(iop->rval);
				ZVAL_BOOL(iop->rval, (iop->rerr == SCERR_BREAK));
				break;

			case SCACHE_IOP_STAT:
			case SCACHE_IOP_SHSTAT:

				/* failed, probably no session. results false */
				MAKE_STD_ZVAL(iop->rval);
				ZVAL_BOOL(iop->rval, 0);
				break;

			case SCACHE_IOP_RENAME:

				MAKE_STD_ZVAL(iop->rval);
				if (iop->rerr == SCERR_SUCCESS) {
					ZVAL_BOOL(iop->rval, 1);
					return 0;
				} else {
					ZVAL_BOOL(iop->rval, 0);
					return 0;
				}

			default:

				zend_error(E_ERROR, 
						   "received unknown response '%x' meaning "
						   "protocol mismatch." 
						   "should not be possible!", restype);
				return -1; /* never */
			}

		} else if ((restype == SCACHE_IOP_STAT) ||
				   (restype == SCACHE_IOP_SHSTAT)) { /* no error */

			/* extract typecode */
			if (rlen < 8) {
				zend_error(E_ERROR,
						   "iov_result: server sent short "
						   "stat packet");
				return -1; /* never */
			}

			switch(sc_ntohl(rbuf + 4)) {
			case SCNODE_EMPTY:

				/* should we treat this as nonexistent 
				MAKE_STD_ZVAL(iop->rval);
				ZVAL_LONG(iop->rval, SCNODE_NONEXISTENT);
				break;
				*/

			case SCNODE_NONEXISTENT:
			case SCNODE_BRANCH:
			case SCNODE_VALUE:

				MAKE_STD_ZVAL(iop->rval);
				ZVAL_LONG(iop->rval, sc_ntohl(rbuf + 4));
				break;

			default:
				
				zend_error(E_ERROR,
						   "SCACHE_IOP_STAT: invalid nodetype %x, "
						   "should not be possible!", 
						   sc_ntohl(rbuf + 4));
				return -1; /* never */
			}

		} else if (restype == SCACHE_IOP_VADD ||
				   restype == SCACHE_IOP_VSUB ||
				   restype == SCACHE_IOP_VGET ||
				   restype == SCACHE_IOP_VSET ||
				   restype == SCACHE_IOP_VUNSET ||
				   restype == SCACHE_IOP_RNSIZE ||
				   restype == SCACHE_IOP_LASTERR) { /* no error */

			/* returns 64bit integer */
			if (rlen < 12) {
				zend_error(E_ERROR,
						   "iov_result: server sent short "
						   "lval packet");
				return -1; /* never */
			}

			MAKE_STD_ZVAL(iop->rval);
			ZVAL_LONG(iop->rval, (int)sc_ntohll(rbuf + 4));

		} else if (restype == SCACHE_IOP_LASTERR) {

			/* returns 32bit integer */
			if (rlen < 8) {
				zend_error(E_ERROR,
						   "iov_result: server sent short "
						   "lasterr packet");
				return -1; /* never */
			}

			MAKE_STD_ZVAL(iop->rval);
			ZVAL_LONG(iop->rval, (int)sc_ntohl(rbuf + 4));

		} else if (restype == SCACHE_IOP_GET ||
				   restype == SCACHE_IOP_SHGET ||
				   restype == SCACHE_IOP_CHGET ||
				   restype == SCACHE_IOP_RNGET ||
				   restype == SCACHE_IOP_RNPOP ||
				   restype == SCACHE_IOP_RNSHIFT) { /* no error */

			/* scache_get() */
			if (reslen > 4 && rlen >= reslen) {
				
				/* first char on resulted string indicates, whether
				   it needs to be unserialized */
				if (rbuf[4] == 'b') {
					
					/* plain string */
					MAKE_STD_ZVAL(iop->rval);
					ZVAL_STRINGL(iop->rval, rbuf + 5,
								 reslen - 5, 1);

				} else if (rbuf[4] == 's') {

					const unsigned char *ptr = (unsigned char *)rbuf + 5;
					php_unserialize_data_t ud;

					MAKE_STD_ZVAL(iop->rval);
					PHP_VAR_UNSERIALIZE_INIT(ud);
					if (!php_var_unserialize(&(iop->rval),
											 &ptr, ptr + reslen - 5, 
											 &ud TSRMLS_CC)) {

						/* should not happen... */
						zend_error(E_ERROR,
								   "SCACHE_IOP_GET: unserialize failed");

						/* never here */
						PHP_VAR_UNSERIALIZE_DESTROY(ud);
						zval_ptr_dtor(&(iop->rval));
						return -1;
					}
					PHP_VAR_UNSERIALIZE_DESTROY(ud);

				} else if (rbuf[4] == 'j') {

					/* java-object, return invalid format */
					sess->scerr = SCERR_INVALID_FORMAT;
					MAKE_STD_ZVAL(iop->rval);
					ZVAL_BOOL(iop->rval, 0);

				} else { /* FIXME. typecode */

					zend_error(E_ERROR,
							   "SCACHE_IOP_GET: invalid typecode, "
							   "should not be possible!");
					return -1; /* never */
				}
			} else {

				zend_error(E_ERROR,
						   "iov_result: response size does not "
						   "match packet header. protocol error.");
				return -1; /* never */
			}

		} else {

			zend_error(E_ERROR,
					   "iov_result: unhandled iop %u (%u), "
					   "should not be here!",
					   restype, reslen);
			return -1; /* never */
		}

		if (!iop->rval) {

			/* should have been terminated */
			zend_error(E_ERROR,
					   "iov_result: success without result, "
					   "shoult not be possible!");
			return -1; /* never */
		}


		/* go for next */
		rlen -= reslen;
		rbuf += reslen;
	}

	if (rlen) {
		zend_error(E_ERROR,
				   "iov_result: respose buffer not fully consumed. "
				   "%i bytes left. possible packet misalignment, "
				   "meaning extension error, not yours", rlen);
		return -1; /* never */
	}

	return 0;
}


/* ==========================================
 * sc_iov_exec
 *
 *  - send bulk request
 * 
 */

static int
sc_iov_exec(struct sc_session_t *sess,
			struct sc_iop_t *iops TSRMLS_DC) {

	int count, tlen;
	char *buffer;

	struct sc_iop_t *iop;
	struct sc_wblock_t *wblock = NULL, **wbf = &wblock;

	/* prepare request-queue and count total length */
	count = tlen = 0;
	for(iop = iops; iop; iop = iop->next, count++) {

		int a1len, a2len;

		/* beware: arg1 ie. path string is always string
		   terminated to first nul byte. you can sen
		   binary key, but only part to first \0 is
		   is used.

		   this is not checked or warned anywhere */
		a1len = strlen(iop->arg1) + 1;
		if ((a2len = iop->arg2_len) == -1) {

			/* protocol-null -exception */
			*wbf = alloca(sizeof(struct sc_wblock_t));
			(*wbf)->next = NULL;
			(*wbf)->length = 12 + a1len;
			(*wbf)->base = buffer = alloca((*wbf)->length + _NMTOBO);

			/* iop */
			buffer = sc_htonl(buffer, 
							  SC_HDR_PACK_R(iop->iop, ((*wbf)->length)));

			buffer = sc_htonl(buffer, (a1len & 0x7fffffff)); /* klen */
			buffer = sc_htonl(buffer, -1); /* vlen */

			/* copy the key */
			memcpy(buffer, iop->arg1, a1len); buffer += a1len;
			
		} else {

			/* kludge the length up, if there's prefix.... */
			if (iop->arg2_prefix) a2len++; 
			
			/* prepare wblock */
			*wbf = alloca(sizeof(struct sc_wblock_t));
			(*wbf)->next = NULL;
			(*wbf)->length = 12 + a1len + a2len;
			(*wbf)->base = buffer = alloca((*wbf)->length + _NMTOBO);

			/* iop */
			buffer = sc_htonl(buffer, 
							  SC_HDR_PACK_R(iop->iop, ((*wbf)->length)));
			buffer = sc_htonl(buffer, (a1len & 0x7fffffff)); /* klen */
			buffer = sc_htonl(buffer, (a2len & 0x7fffffff)); /* vlen */
			
			/* copy the key */
			memcpy(buffer, iop->arg1, a1len); buffer += a1len;
			
			/* add prefix-byte if exists */
			if (iop->arg2_prefix) {
				*buffer++ = iop->arg2_prefix;
				
				/* ...and kludge the length back down */
				a2len--;
			}

			/* and copy the data */
			memcpy(buffer, iop->arg2, a2len); buffer += a2len;
		}

		/* go for next */
		tlen += (*wbf)->length;
		wbf = &((*wbf)->next);
	}

	/* DEPENDS: sc_query (PROTOHEADER) 
	 * 
	 * sc_query constructs main header */

	/* prepend whole queue with main header */
	sess->header->next = wblock;
	wblock = sess->header;
	tlen += wblock->length;

	buffer = wblock->base;
	buffer = sc_htonl(buffer, SC_HDR_PACK_H(count, tlen));	
	buffer = sc_htonl(buffer, sess->index);
	buffer = sc_htonl(buffer, 0);

	/* send... */
	sess->scerr = sc_post(sess, wblock, max_retries TSRMLS_CC);

	if (sess->scerr == SCERR_SUCCESS) {

		char *stackbuf;
		struct sc_req_t req;

		req.flags = 0;
		req.rlen = 32768;
		req.rbuffer = stackbuf = alloca(req.rlen + _NMTOBO);

		sess->scerr = sc_read(sess, &req, count TSRMLS_CC);
		if (sess->scerr == SCERR_SUCCESS) {

			/* actually there is only success,
			   all errors zend_error() */

			int rv = sc_iov_result(sess, &req, iops TSRMLS_CC);
			if (req.rbuffer != stackbuf) {
				/* got replaced */
				sc_efree(req.rbuffer);
			}

			/* same with sc_iov_result, always success. */
			return rv;
		}
		
		zend_error(E_ERROR,
				   "reading reply from server failed. "
				   "server code is %s\n", 
				   sc_strerror(sess->scerr)); 
	}

	zend_error(E_ERROR,
			   "querying server failed. "
			   "server code is %s\n", 
			   sc_strerror(sess->scerr)); 
	return -1; /* never */
}


/* =================================================
 * sc_iov
 *
 *  - bulk request 
 *
 */
static void
sc_iov_r(INTERNAL_FUNCTION_PARAMETERS, char *funcname, 
		 struct sc_session_t *sess, HashTable *ahash) {

	int iop_count = 0;
	zval **oparr;
	HashPosition pos;

	struct sc_iop_t *iops = NULL;
	struct sc_iop_t **iopsp = &iops;
	
	zend_hash_internal_pointer_reset_ex(ahash, &pos);
	while(zend_hash_get_current_data_ex
		  (ahash, (void **)&oparr, &pos) == SUCCESS) {
		
		HashTable *row;
		HashPosition rpos;
		
		if (Z_TYPE_PP(oparr) != IS_ARRAY) {
			zend_error(E_WARNING, "%s: invalid arguments", funcname);
			RETURN_FALSE;
		}
		
		row = Z_ARRVAL_PP(oparr);
		zend_hash_internal_pointer_reset_ex(row, &rpos);
		
		*iopsp = sc_iov_extract_iop(row, &rpos, 
									alloca(sizeof(struct sc_iop_t)));
		if (*iopsp) {
			
			iopsp = &((*iopsp)->next);
			iop_count++;
			
		} else {
			
			zend_error(E_WARNING, "%s: invalid IOP parameters", funcname);
			RETURN_FALSE;
		}
		
		zend_hash_move_forward_ex(ahash, &pos);
	}

	if (iops) {

		if (iop_count > 0x7f) { /* artificial...?? why??? */

			zend_error(E_ERROR,
					   "%s: too many requests. "
					   "got %i when max is 127", 
					   funcname, iop_count);
			/* die */

		} else if (sc_iov_exec(sess, iops TSRMLS_CC) < 0) {

			/* should not happen, sc_iov_exec dies internally */
			zend_error(E_ERROR,
					   "%s: IOP exec failed, "
					   "extension's error, not yours", funcname);

		} else {

			/* protocol ok, construct result array from
			   responses */
			array_init(return_value);
			while(iops) {
				zend_hash_next_index_insert(HASH_OF(return_value),
											&(iops->rval), 
											sizeof(zval *), NULL);

				iops = iops->next;
			}

			return;
		}
	}

	/* no request */
	zend_error(E_WARNING, 
			   "%s: nothing to request. " 
			   "request array was invalid", funcname);
	RETURN_FALSE;
}

static void
sc_iov(INTERNAL_FUNCTION_PARAMETERS) {

	int argc;
	zval ***argv;
	struct sc_session_t *sess;

	argc = ZEND_NUM_ARGS();
	if (argc != 2) WRONG_PARAM_COUNT; /* (< 2) for non-array... */

	argv = alloca(argc * sizeof(argv[0]));
	if (zend_get_parameters_array_ex(argc, argv) != SUCCESS) {
		WRONG_PARAM_COUNT;
		/* dies */
	}

	/* session */
	if (Z_TYPE_P(*argv[0]) == IS_RESOURCE) {

		sess = (struct sc_session_t *)
			zend_fetch_resource(argv[0] , -1, SC_SESS_RESNAME, NULL, 1,
								le_sessions);
		if (!sess) {
			zend_error(E_WARNING, "scache_iov: no session");
			RETURN_FALSE;
		}

	} else {
		zend_error(E_WARNING, 
				   "scache_iov expects parameter 1 to be resource");
		RETURN_FALSE;
	}

	/* array supplied */
	if (argc == 2 && Z_TYPE_P(*argv[1]) == IS_ARRAY) {
		return sc_iov_r(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
						"scache_iov", sess,	Z_ARRVAL_P(*argv[1]));
	} else { /* list specified */

		/* TODO, iop-list as parameters */
		zend_error(E_WARNING, "sc_iov: unimplemented");
		WRONG_PARAM_COUNT;
	}

}

static void
sc_iov_m(INTERNAL_FUNCTION_PARAMETERS) {

	int argc;
	zval ***argv;
	struct sc_classdata_t *cdata = (struct sc_classdata_t *)
		zend_object_store_get_object(getThis() TSRMLS_CC);

	argc = ZEND_NUM_ARGS();
	if (argc != 1) WRONG_PARAM_COUNT; /* (< 1) for non-array... */

	argv = alloca(argc * sizeof(argv[0]));
	if (zend_get_parameters_array_ex(argc, argv) != SUCCESS) {
		WRONG_PARAM_COUNT;
	}

	if (Z_TYPE_P(*argv[0]) == IS_ARRAY) {
		return sc_iov_r(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
						"SCache->iov", cdata->session, 
						Z_ARRVAL_P(*argv[0]));
	} else {
		
		/* TODO, iop-list as parameters */
		zend_error(E_WARNING, "SCache->iov: unimplemented");
		WRONG_PARAM_COUNT;
	}
}


/* ====================================
 * sc_iop_single
 *
 *  - mainly wrapper for single-operation
 *    php functions
 *
 */

static void
sc_iop_single(INTERNAL_FUNCTION_PARAMETERS, 
			  char *funcname, int argcount, int op) {

	zval *zv, *kv, *pv;

	struct sc_iop_t iop;
	struct sc_session_t *sess;

	memset(&iop, 0, sizeof(iop));
	iop.iop = op;
	iop.arg2 = "";

	/* called through class interface? */
	if (argcount == 3) {
		if (!getThis()) {
			if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
									  "rzz", &zv, &kv, 
									  &pv) == FAILURE) RETURN_FALSE;
		} else {
			if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
									  "zz", &kv, 
									  &pv) == FAILURE) RETURN_FALSE;
		}
			
		if (sc_serialize_arg3("scache_iov", pv, &iop)) RETURN_FALSE;

	} else { /* argcount == 2 */
		if (!getThis()) {
			if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
									  "rz", &zv, &kv) == FAILURE) 
				RETURN_FALSE;
		} else {
			if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
									  "z", &kv) == FAILURE) 
				RETURN_FALSE;
		}
	}

	/* keypath */
	if (sc_verify_as_key2(funcname, kv) < 0) RETURN_FALSE;
	iop.arg1 = Z_STRVAL_P(kv);


	/* session */
	if (!getThis()) {

		sess = (struct sc_session_t *)
			zend_fetch_resource(&zv , -1, SC_SESS_RESNAME, NULL, 1,
								le_sessions);
	} else { /* class-interface */

		struct sc_classdata_t *cdata = (struct sc_classdata_t *)
			zend_object_store_get_object(getThis() TSRMLS_CC);
		sess = cdata->session;
	}

	if (sess) {

		if (sc_iov_exec(sess, &iop TSRMLS_CC) < 0) {
			
			/* should not happen, sc_iov_exec dies internally */
			zend_error(E_ERROR,
					   "%s: sc_iop_single failed: "
					   "extension's error, not yours", funcname);
			
		} else {

			*return_value = *iop.rval;
			zval_copy_ctor(return_value);
			zval_ptr_dtor(&(iop.rval));
			return;
		}
		
	} else {
		
		zend_error(E_ERROR, "%s: no session", funcname);
	}

	/* else */
	RETURN_FALSE;
}

static void
sc_iop_coll(INTERNAL_FUNCTION_PARAMETERS,
			char *funcname, int op) {

	zval *zv, *kv, *pv;

	struct sc_iop_t iop;
	struct sc_session_t *sess;

	memset(&iop, 0, sizeof(iop));
	iop.iop = op;
	iop.arg2 = "";

	if (!getThis()) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
								  "rzz", &zv, &kv, 
								  &pv) == FAILURE) RETURN_FALSE;
	} else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
								  "zz", &kv, 
								  &pv) == FAILURE) RETURN_FALSE;
	}
	
	/* both are keys */
	if (sc_verify_as_key2(funcname, kv) < 0) RETURN_FALSE;
	iop.arg1 = Z_STRVAL_P(kv);
	if (sc_verify_as_key3(funcname, pv) < 0) RETURN_FALSE;
	iop.arg2 = Z_STRVAL_P(pv);
	iop.arg2_len = Z_STRLEN_P(pv) + 1;
	iop.arg2_prefix = 0;

	/* session */
	if (!getThis()) {
		sess = (struct sc_session_t *)
			zend_fetch_resource(&zv , -1, SC_SESS_RESNAME, NULL, 1,
								le_sessions);
	} else { /* class-interface */

		struct sc_classdata_t *cdata = (struct sc_classdata_t *)
			zend_object_store_get_object(getThis() TSRMLS_CC);
		sess = cdata->session;
	}

	if (sess) {

		if (sc_iov_exec(sess, &iop TSRMLS_CC) < 0) {
			
			/* should not happen, sc_iov_exec dies internally */
			zend_error(E_ERROR,
					   "%s: sc_iop_single failed: "
					   "extension's error, not yours", funcname);
			
		} else {

			*return_value = *iop.rval;
			zval_copy_ctor(return_value);
			zval_ptr_dtor(&(iop.rval));
			return;
		}
		
	} else {
		
		zend_error(E_ERROR, "%s: no session", funcname);
	}

	/* else */
	RETURN_FALSE;
}

/* ====================================
 * sc_iop_lval
 *
 *  - equal to sc_iop_single, but with lval arg
 *  - mainly wrapper for single-operation
 *    php functions
 *
 */
static void
sc_iop_lval(INTERNAL_FUNCTION_PARAMETERS,
			char *funcname, int op) {

	long value = 1;
	zval *zv, *kv;

	struct sc_session_t *sess;

	if (!getThis()) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
								  "rz|l", &zv, &kv, &value) == FAILURE) {
			RETURN_FALSE;
		}
	} else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
								  "z|l", &kv, &value) == FAILURE) {
			RETURN_FALSE;
		}
	}

	/* keypath */
	if (sc_verify_as_key2(funcname, kv) < 0) {
		RETURN_FALSE;
	}

	/* session */
	if (!getThis()) {
		sess = (struct sc_session_t *)
			zend_fetch_resource(&zv , -1, SC_SESS_RESNAME, NULL, 1,
							le_sessions);
	} else { /* class-interface */

		struct sc_classdata_t *cdata = (struct sc_classdata_t *)
			zend_object_store_get_object(getThis() TSRMLS_CC);
		sess = cdata->session;
	}

	if (sess) {

		char vbuf[8]; /* for int64_t */
		struct sc_iop_t iop;

		memset(&iop, 0, sizeof(iop));

		iop.iop = op;
		iop.arg1 = Z_STRVAL_P(kv);
		iop.arg2 = vbuf;
		iop.arg2_len = 8;
		sc_htonll(vbuf, (int64_t)value);

		if (sc_iov_exec(sess, &iop TSRMLS_CC) < 0) {

			/* should not happen, sc_iov_exec dies internally */
			zend_error(E_ERROR,
					   "%s: sc_iop_single failed: "
					   "extension's error, not yours", funcname);
			
		} else {

			*return_value = *iop.rval;
			zval_copy_ctor(return_value);
			zval_ptr_dtor(&(iop.rval));
			return;
		}
		
	} else {
		zend_error(E_ERROR, "%s: no session", funcname);
	}

	/* else */
	RETURN_FALSE;
}

/* ===================================
 * sc_exists_wrapper
 *
 *  - convert packet integer to php boolean
 *  - wrapper
 *
 */

static void
sc_exists_wrapper(INTERNAL_FUNCTION_PARAMETERS, 
				  char *funcname, int op) {

	/* call */
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  funcname, 2, op);

	/* mangle to boolean */
	if (return_value->type == IS_LONG) {

		switch(return_value->value.lval) {
		case SCNODE_EMPTY:
		case SCNODE_NONEXISTENT:
			return_value->type = IS_BOOL;
			return_value->value.lval = 0;
			break;
		case SCNODE_BRANCH:
		case SCNODE_VALUE:
			return_value->type = IS_BOOL;
			return_value->value.lval = 1;
			break;
		default:
			zend_error(E_ERROR,
					   "%s: unknown nodetype %lu",
					   funcname, return_value->value.lval);

			return_value->type = IS_BOOL; /* never */
			return_value->value.lval = 0;
			break;
		}

	} else {

		zend_error(E_ERROR,
				   "%s extension failed: (not script) "
				   "wrong type of result", funcname);
	}
}

static void
sc_rename(INTERNAL_FUNCTION_PARAMETERS, char *funcname) {

	int len;
	char *tag;
	zval *zv, *nv;

	struct sc_iop_t iop;
	struct sc_session_t *sess;

	memset(&iop, 0, sizeof(iop));
	iop.iop = SCREQ_TYPE_SESSION_RENAME;
	iop.arg2 = "";

	if (!getThis()) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
								  "rz", &zv, &nv) == FAILURE) {
			RETURN_FALSE;
		}
		
		/* session */
		sess = (struct sc_session_t *)
			zend_fetch_resource(&zv , -1, SC_SESS_RESNAME, NULL, 1,
								le_sessions);

		if (!sess) {
			zend_error(E_WARNING, "%s: no session", funcname);
			RETURN_FALSE;
		}

	} else { /* class-interface */

		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
								  "z", &nv) == FAILURE) {
			RETURN_FALSE;
		}

		struct sc_classdata_t *cdata = (struct sc_classdata_t *)
			zend_object_store_get_object(getThis() TSRMLS_CC);
		sess = cdata->session;
	}

	/* verify new session identifier */
	if ((Z_TYPE_P(nv) != IS_STRING) || (Z_STRVAL_P(nv) == NULL)) {

		zend_error(E_WARNING,
				   "%s expexts parameter 2 to be non-null "
				   "string", funcname);
		RETURN_FALSE;
	}

	len = strlen(Z_STRVAL_P(nv));
	if (len < SCACHE_PTO_MIN_TAGLEN) {
		zend_error(E_WARNING,
				   "%s: new session identifier specified "
				   "is empty", funcname);
		RETURN_FALSE;
	} else if (len != Z_STRLEN_P(nv)) {
		zend_error(E_WARNING,
				   "%s: key has nul bytes in middle. key "
				   "needs to be plain nul-terminated string, not "
				   "binary string.", funcname);
		RETURN_FALSE;
	}

	/* append tag */
	tag = sc_remap_tag(Z_STRVAL_P(nv));
	if (strlen(tag) > SCACHE_PTO_MAX_TAGLEN) {
		zend_error(E_WARNING,
				   "%s: after remapping session identifier "
				   "is too long", funcname);
		sc_pefree(tag, 1);
		RETURN_FALSE;
	}

	iop.arg1 = tag;
	if (sc_iov_exec(sess, &iop TSRMLS_CC) < 0) {
		
		/* should not happen, sc_iov_exec dies internally */
		zend_error(E_ERROR,
				   "%s: sc_iov_exec() failed: "
				   "extension's error, not yours", funcname);
		sc_pefree(tag, 1);
		RETURN_FALSE;
	}

	/* result */
	if (iop.rval->type != IS_BOOL) {
		
		/* protocol error */
		zend_error(E_ERROR,
				   "%s: backend returned non-boolean "
				   "value. this is bad thing...", funcname);

		/* TODO: errcode */
		sc_pefree(tag, 1);
		RETURN_FALSE;
	}

	*return_value = *iop.rval;
	zval_copy_ctor(return_value);
	zval_ptr_dtor(&(iop.rval));

	return_value->type = IS_BOOL;
	return_value->value.lval = 
		(return_value->value.lval) ? 1 : 0;

	if (return_value->value.lval) {

		int taglen, seclen, athlen;
		char *auth, *buffer;

		/* auth-string */
		auth = SCACHE_G(ini_auth_string);
		if (!auth) auth = "";

		/* replace session identifier in current session.. */
		sc_pefree(sess->tag, 1);
		sess->tag = tag;

		taglen = strlen(sess->tag) + 1;
		seclen = strlen(sess->secret) + 1;
		athlen = strlen(auth) + 1;

		/* ...and recreate headers */
		sc_pefree(sess->header->base, 1);
		sc_pefree(sess->header, 1);

		sess->header = sc_pemalloc(sizeof(struct sc_wblock_t), 1);
		sess->header->length = 18 + taglen + seclen + athlen;
		sess->header->base = buffer = 
			sc_pemalloc(sess->header->length + _NMTOBO, 1);

		/* ...and prepare parts that are static */
		buffer = sc_htons(buffer + 12, taglen);
		buffer = sc_htons(buffer, seclen);
		buffer = sc_htons(buffer, athlen);
		
		memcpy(buffer, sess->tag, taglen);
		buffer += taglen;
		
		memcpy(buffer, sess->secret, seclen);
		buffer += seclen;
		
		memcpy(buffer, auth, athlen);
		buffer += seclen;
	}

	return;
}

static void
sc_drop(INTERNAL_FUNCTION_PARAMETERS, char *funcname) {

	char *stackbuf;
	struct sc_req_t req;

	req.reqtype = SCREQ_TYPE_SESSION_DROP;
	req.flags = 0;
	req.rlen = 256;
	req.rbuffer = stackbuf = alloca(req.rlen);
	
	if (sc_simple_command(INTERNAL_FUNCTION_PARAM_PASSTHRU,
						  funcname, &req) == SCERR_SUCCESS) {

		/* TODO: destroy session-resource */

		RETVAL_BOOL(1);

	} else {
		
		RETVAL_BOOL(0);
	}

	/* whether stackbuf got replaced */
	if (req.rbuffer != stackbuf) sc_efree(req.rbuffer);
}


static void
sc_opt(INTERNAL_FUNCTION_PARAMETERS, const char *funcname, int op) {

	char *buffer, *stackbuf;
	struct sc_req_t req;
	struct sc_session_t *sess;

	long opt,val;
	zval *zv;

	req.reqtype = op;
	req.flags = 0;
	req.rlen = 64;
	req.rbuffer = stackbuf = alloca(req.rlen);

	if (op == SCACHE_IOP_SETOPT) {

		if (!getThis()) {
			if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
									  "rll", &zv, &opt, &val) == FAILURE) {
				RETURN_FALSE;
			}
		} else { /* class-interface */
			if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
									  "ll", &opt, &val) == FAILURE) {
				RETURN_FALSE;
			}
		}

		req.wblocks = alloca(sizeof(struct sc_wblock_t));
		req.wblocks->length = 8;
		req.wblocks->base = buffer = alloca(req.wblocks->length);
		req.wblocks->next = NULL;

		buffer = sc_htonl(buffer, opt);
		buffer = sc_htonl(buffer, val);

	} else { /* SCACHE_IOP_GETOPT */

		if (!getThis()) {
			if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
									  "rl", &zv, &opt) == FAILURE) {
				RETURN_FALSE;
			}
		} else {
			if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
									  "l", &opt) == FAILURE) {
				RETURN_FALSE;
			}
		}

		req.wblocks = alloca(sizeof(struct sc_wblock_t) + _NMTOBO);
		req.wblocks->length = 4;
		req.wblocks->base = buffer = alloca(req.wblocks->length + _NMTOBO);
		req.wblocks->next = NULL;

		buffer = sc_htonl(buffer, opt);
	}

	RETVAL_BOOL(0);

	/* session */
	if (!getThis()) {
		sess = (struct sc_session_t *)
			zend_fetch_resource(&zv, -1, SC_SESS_RESNAME, NULL, 1,
								le_sessions);
	} else {
		struct sc_classdata_t *cdata = (struct sc_classdata_t *)
			zend_object_store_get_object(getThis() TSRMLS_CC);
		sess = cdata->session;
	}

	if (sess) {

		int rlen;
		char *rbuf;

		if ((sc_query(sess, &req TSRMLS_CC) == SCERR_SUCCESS)) {

			sess->scerr =
				sc_parse_simple_response(&req, 
										 &rlen, &rbuf);

			if (sess->scerr == SCERR_SUCCESS) {

				/* success returns old value as int32 */
				if (rlen == 4) {
					
					switch(opt) {
					case SCACHE_OPT_EXPIRES:
						RETVAL_LONG(sc_ntohl(rbuf));
						break;
					case SCACHE_OPT_KEEPONELIMIT:
						RETVAL_BOOL(!!sc_ntohl(rbuf));
						break;
					default:
						zend_error(E_ERROR, "%s: unimplemented option %li",
								   funcname, opt);
						break;
					}
				} else {
					zend_error(E_ERROR, "%s: got short response", funcname);
				}
			} else {
				RETVAL_BOOL(0);
			}

		} else {
			
			/* should not happen */
			zend_error(E_ERROR, "%s: operation failed", funcname);
		}

	} else {
		zend_error(E_WARNING, "%s: no session", funcname);
	}

	if (req.rbuffer != stackbuf) {

		/* got replaces */
		sc_efree(req.rbuffer);
	}
}

static void
sc_status(INTERNAL_FUNCTION_PARAMETERS, char *funcname) {

	char *stackbuf;
	struct sc_req_t req;
	enum scache_error_t scerr;

	req.reqtype = SCREQ_TYPE_NONE;
	req.flags = 0;
	req.rlen = 128;
	req.rbuffer = stackbuf = alloca(req.rlen);

	scerr = sc_simple_command(INTERNAL_FUNCTION_PARAM_PASSTHRU,
							  funcname, &req);

	if (scerr == SCERR_SUCCESS) {
		RETVAL_BOOL(1);
	} else {
		RETVAL_BOOL(0);
	}

	if (req.rbuffer != stackbuf) sc_efree(req.rbuffer);
	return;
}

static void
sc_srvstats(INTERNAL_FUNCTION_PARAMETERS, char *funcname) {

	char *stackbuf;
	struct sc_req_t req;
	enum scache_error_t scerr;

	req.reqtype = SCREQ_TYPE_SRVSTATS;
	req.flags = 0;
	req.rlen = 8192;
	req.rbuffer = stackbuf = alloca(req.rlen);

	RETVAL_BOOL(0);
	scerr = sc_simple_command(INTERNAL_FUNCTION_PARAM_PASSTHRU,
							  funcname, &req);
	if (scerr == SCERR_SUCCESS) {

		/* unserialize */
		zval *zv;
		const unsigned char *ptr = 
			(unsigned char *)req.rbuffer + 4;
		php_unserialize_data_t ud;
		
		
		if (req.rlen < 5) { /* TODO: efree(stackbuf), reorder */

			if (req.rbuffer != stackbuf) sc_efree(req.rbuffer);

			zend_error(E_ERROR, "%s: got short response", funcname);
			RETURN_FALSE; /* never */ 
		}
		
		
		MAKE_STD_ZVAL(zv);
		PHP_VAR_UNSERIALIZE_INIT(ud);
		if (!php_var_unserialize(&zv, &ptr, ptr + req.rlen - 4, 
								 &ud TSRMLS_CC) ||
			Z_TYPE_P(zv) != IS_ARRAY) {
			
			/* should not happen... */
			zend_error(E_WARNING,
					   "%s: unserialize failed.", funcname);
			
		} else {
			
			*return_value = *zv;
			zval_copy_ctor(return_value);
		}
		
		PHP_VAR_UNSERIALIZE_DESTROY(ud);
		zval_ptr_dtor(&zv);
		
	} else if (scerr == SCERR_NO_ACCESS) {
		
		zend_error(E_WARNING, "%s: " 
				   "status queries not permitted by server", funcname);

	} else {
		
		zend_error(E_WARNING, "%s: "
				   "should not be here, got error %x", 
				   funcname, scerr);
	}
}


static void
sc_getindex(INTERNAL_FUNCTION_PARAMETERS) {

	zval *zv;
	struct sc_session_t *sess;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
							  "r", &zv) == FAILURE) {

		RETURN_FALSE;
	}

	sess = (struct sc_session_t *)
		zend_fetch_resource(&zv , -1, SC_SESS_RESNAME, NULL, 1,
							le_sessions);
	if (sess) {
		RETURN_LONG(sess->index);
	} else {
		zend_error(E_WARNING, "sc_getindex(): no session");
		RETURN_FALSE;
	}
}

static void
sc_getindex_m(INTERNAL_FUNCTION_PARAMETERS) {

	struct sc_classdata_t *cdata = (struct sc_classdata_t *)
		zend_object_store_get_object(getThis() TSRMLS_CC);

	if (ZEND_NUM_ARGS() > 0) {
		WRONG_PARAM_COUNT;
	} else {
		RETURN_LONG(cdata->session->index);
	}
}

static void
sc_lasterr(INTERNAL_FUNCTION_PARAMETERS) {

	zval *zv;
	struct sc_session_t *sess;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
							  "r", &zv) == FAILURE) {

		RETURN_FALSE;
	}

	sess = (struct sc_session_t *)
		zend_fetch_resource(&zv , -1, SC_SESS_RESNAME, NULL, 1,
							le_sessions);
	if (sess) {
		RETURN_LONG(sess->scerr);
	} else {
		zend_error(E_WARNING, "sc_lasterr(): no session");
		RETURN_FALSE;
	}
}

static void
sc_lasterr_m(INTERNAL_FUNCTION_PARAMETERS) {

	struct sc_classdata_t *cdata = (struct sc_classdata_t *)
		zend_object_store_get_object(getThis() TSRMLS_CC);

	if (ZEND_NUM_ARGS() > 0) {
		WRONG_PARAM_COUNT;
	} else {
		RETURN_LONG(cdata->session->scerr);
	}
}

static void
sc_map_strerror(INTERNAL_FUNCTION_PARAMETERS) {

	long scerr;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
							  "l", &scerr) == FAILURE) {
		RETURN_FALSE;
	}

	RETURN_STRING(sc_strerror(scerr), 1);
}

static void
sc_map_strerror_m(INTERNAL_FUNCTION_PARAMETERS) {

	long scerr = -1;
	struct sc_classdata_t *cdata = (struct sc_classdata_t *)
		zend_object_store_get_object(getThis() TSRMLS_CC);

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
							  "|l", &scerr) == FAILURE) {
		RETURN_FALSE;
	}

	if (scerr < 0) scerr = cdata->session->scerr;
	RETURN_STRING(sc_strerror(scerr), 1);
}


#ifdef HAVE_PHP_SESSION
/* ==================================
 * session-handler
 * ================================== */
static struct sc_session_t *
sc_get_session_mod_data(void **mod_data TSRMLS_DC) {

	int port;
	const char *host;
	struct sc_conn_t *conn;
	struct sc_session_t *sess;

	/* already */
	if (!*mod_data) *mod_data = sc_connect_real(-1, NULL, "_" /* dummy */,
												NULL, 0, 1);
	return *mod_data;


	/* no, create */
	*mod_data = sc_connect_real(-1, NULL, "_" /* dummy */,
								NULL, 0, 1 TSRMLS_CC);
	return *mod_data;

	host = SCACHE_G(ini_default_host);
	if (!host) {
		host = SCACHE_PTO_UNIX_SOCK;
	} else {
		port = SCACHE_G(ini_default_port);
		if (port < 1 || port > 65535) port = SCACHE_PTO_LISTEN_PORT;
	}

	conn = sc_conn_create(host, port, 1 TSRMLS_CC);
	if (conn) {

		sess = sc_session_alloc(conn, "", "");
		if (sess) {
			PS_SET_MOD_DATA((void *)sess);
			return sess;
		}
	}

	/* fail */
	return NULL;
}

PS_OPEN_FUNC(scache) {


	return sc_get_session_mod_data(mod_data TSRMLS_CC) ?
		SUCCESS : FAILURE;
}

PS_READ_FUNC(scache) {

	char *stackbuf, *buffer;
	struct sc_req_t req;
	struct sc_session_t *sess;
	
	sess = sc_get_session_mod_data(mod_data TSRMLS_CC);
	if (!sess) return FAILURE;

	/* request */
	req.reqtype = SCREQ_TYPE_SESGET;
	req.flags = 0;
	req.rlen = 32768;
	req.rbuffer = stackbuf = alloca(req.rlen);

	/* key */
	req.wblocks = alloca(sizeof(struct sc_wblock_t));
	req.wblocks->length = strlen(key) + 1 + 8;
	req.wblocks->base = buffer = alloca(req.wblocks->length + _NMTOBO);
	req.wblocks->next = NULL;

	buffer = sc_htonl(buffer, INI_INT("session.gc_maxlifetime"));
	buffer = sc_htonl(buffer, strlen(key) + 1);
	strcpy(buffer, key);
	
	if (sc_query(sess, &req TSRMLS_CC) == SCERR_SUCCESS) {

		if (req.rlen < 5) {

			zend_error(E_WARNING, "session_read(): got short response");
			goto return_fail;
		}

		/* else */
		*vallen = req.rlen - 4;
		*val = sc_emalloc(req.rlen /* - 4 (NMTOBO) */);
		memcpy(*val, req.rbuffer + 4, *vallen);

		if (req.rbuffer != stackbuf) sc_efree(req.rbuffer);
		return SUCCESS;
	}

 return_fail:

	*val = sc_emalloc(1);
	(*val)[0] = 0;
	*vallen = 0;

	if (req.rbuffer != stackbuf) sc_efree(req.rbuffer);
	return FAILURE;
}

PS_WRITE_FUNC(scache) {

	int rlen;
	char *stackbuf, *buffer;
	struct sc_req_t req;
	struct sc_session_t *sess;

	/* skip empty writes */
	if (!(val && *val)) return SUCCESS;

	sess = sc_get_session_mod_data(mod_data TSRMLS_CC);
	if (!sess) return FAILURE;

	/* request */
	req.reqtype = SCREQ_TYPE_SESPUT;
	req.flags = 0;
	req.rlen = 128;
	req.rbuffer = stackbuf = alloca(req.rlen);

	/* expires + key */
	req.wblocks = alloca(sizeof(struct sc_wblock_t));
	req.wblocks->length = strlen(key) + 1 + 12;
	req.wblocks->base = buffer = alloca(req.wblocks->length + _NMTOBO);

	buffer = sc_htonl(buffer, INI_INT("session.gc_maxlifetime"));
	buffer = sc_htonl(buffer, strlen(key) + 1);
	buffer = sc_htonl(buffer, vallen);
	strcpy(buffer, key);

	/* store-block */
	req.wblocks->next = alloca(sizeof(struct sc_wblock_t));
	req.wblocks->next->length = vallen;
	req.wblocks->next->base = (char *)val;
	req.wblocks->next->next = NULL;

	if (sc_query(sess, &req TSRMLS_CC) == SCERR_SUCCESS) {
		sess->scerr = sc_parse_simple_response(&req, &rlen, &buffer);
	}

	if (req.rbuffer != stackbuf) sc_efree(req.rbuffer);
	return sess->scerr == SCERR_SUCCESS ? SUCCESS : FAILURE;
}

PS_CLOSE_FUNC(scache) {

	struct sc_session_t *sess = (struct sc_session_t *)*mod_data;

	if (sess) { 
		sc_session_clear(sess);
		*mod_data = NULL;
	}

	return SUCCESS;
}

PS_DESTROY_FUNC(scache) {

	char *stackbuf, *buffer;
	struct sc_req_t req;
	struct sc_session_t *sess;

	sess = sc_get_session_mod_data(mod_data TSRMLS_CC);
	if (!sess) return FAILURE;

	/* request */
	req.reqtype = SCREQ_TYPE_SESDROP;
	req.flags = 0;
	req.rlen = 32768;
	req.rbuffer = stackbuf = alloca(req.rlen);

	/* key */
	req.wblocks = alloca(sizeof(struct sc_wblock_t));
	req.wblocks->length = strlen(key) + 1 + 4;
	req.wblocks->base = buffer = alloca(req.wblocks->length + _NMTOBO);
	req.wblocks->next = NULL;

	buffer = sc_htonl(buffer, strlen(key) + 1);
	strcpy(buffer, key);
	
	sc_query(sess, &req TSRMLS_CC);
	if (req.rbuffer != stackbuf) sc_efree(req.rbuffer);

	return (sess->scerr == SCERR_SUCCESS) ? SUCCESS : FAILURE;
}

PS_GC_FUNC(scache) {

	/* nothing to do */
	return SUCCESS;
}
#endif /* HAVE_PHP_SESSION */




/* ==================================
 * module wrappers
 * ================================== */
ZEND_FUNCTION(scache_open) {
	sc_open(INTERNAL_FUNCTION_PARAM_PASSTHRU, persist_off, 0); }
ZEND_FUNCTION(scache_popen) {
	sc_open(INTERNAL_FUNCTION_PARAM_PASSTHRU, persist_on, 0); }
ZEND_FUNCTION(scache_reset) {
	sc_open(INTERNAL_FUNCTION_PARAM_PASSTHRU, persist_off, 1); }
ZEND_FUNCTION(scache_preset) {
	sc_open(INTERNAL_FUNCTION_PARAM_PASSTHRU, persist_on, 1); }
ZEND_FUNCTION(scache_connect) {
	sc_connect(INTERNAL_FUNCTION_PARAM_PASSTHRU, persist_off); }
ZEND_FUNCTION(scache_pconnect) {
	sc_connect(INTERNAL_FUNCTION_PARAM_PASSTHRU, persist_on); }
ZEND_FUNCTION(scache_connect_fast) {
	sc_connect_fast(INTERNAL_FUNCTION_PARAM_PASSTHRU, persist_off); }
ZEND_FUNCTION(scache_pconnect_fast) {
	sc_connect_fast(INTERNAL_FUNCTION_PARAM_PASSTHRU, persist_on); }
ZEND_FUNCTION(scache_rename) {
	sc_rename(INTERNAL_FUNCTION_PARAM_PASSTHRU, "sache_rename"); }
ZEND_FUNCTION(scache_drop) {
	sc_drop(INTERNAL_FUNCTION_PARAM_PASSTHRU, "scache_drop"); }


ZEND_FUNCTION(scache_getopt) {
	sc_opt(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
		   "scache_getopt", SCACHE_IOP_GETOPT); }
ZEND_FUNCTION(scache_setopt) {
	sc_opt(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
		   "scache_setopt", SCACHE_IOP_SETOPT); }
ZEND_FUNCTION(scache_status) {
	sc_status(INTERNAL_FUNCTION_PARAM_PASSTHRU, "scache_status"); }
ZEND_FUNCTION(scache_srvstats) {
	sc_srvstats(INTERNAL_FUNCTION_PARAM_PASSTHRU, "scache_srvstats"); }
ZEND_FUNCTION(scache_getindex) {
	sc_getindex(INTERNAL_FUNCTION_PARAM_PASSTHRU); }
ZEND_FUNCTION(scache_lasterr) {
	sc_lasterr(INTERNAL_FUNCTION_PARAM_PASSTHRU); }
ZEND_FUNCTION(scache_strerror) {
	sc_map_strerror(INTERNAL_FUNCTION_PARAM_PASSTHRU); }
ZEND_FUNCTION(scache_iov) {
	sc_iov(INTERNAL_FUNCTION_PARAM_PASSTHRU); }


/* one-arg iop-wrappers */
ZEND_FUNCTION(scache_get) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "scache_get", 2, SCACHE_IOP_GET); }
ZEND_FUNCTION(scache_unset) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "scache_unset", 2, SCACHE_IOP_UNSET); }
ZEND_FUNCTION(scache_stat) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "scache_stat", 2, SCACHE_IOP_STAT); }
ZEND_FUNCTION(scache_exists) {
	sc_exists_wrapper(INTERNAL_FUNCTION_PARAM_PASSTHRU,
					  "scache_exists", SCACHE_IOP_STAT); }


ZEND_FUNCTION(scache_shget) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "scache_shget", 2, SCACHE_IOP_SHGET); }
ZEND_FUNCTION(scache_shunset) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "scache_shunset", 2, SCACHE_IOP_SHUNSET); }
ZEND_FUNCTION(scache_shstat) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "scache_shstat", 2, SCACHE_IOP_SHSTAT); }
ZEND_FUNCTION(scache_shexists) {
	sc_exists_wrapper(INTERNAL_FUNCTION_PARAM_PASSTHRU,
					  "scache_shexists", SCACHE_IOP_SHSTAT); }


ZEND_FUNCTION(scache_chget) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "scache_chget", 2, SCACHE_IOP_CHGET); }
ZEND_FUNCTION(scache_chclear) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "scache_chclear", 2, SCACHE_IOP_CHCLEAR); }

ZEND_FUNCTION(scache_vget) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU,
				  "scache_vget", 2, SCACHE_IOP_VGET); }
ZEND_FUNCTION(scache_vunset) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU,
				  "scache_vunset", 2, SCACHE_IOP_VUNSET); }

ZEND_FUNCTION(scache_rnget) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU,
				  "scache_rnget", 2, SCACHE_IOP_RNGET); }
ZEND_FUNCTION(scache_rnsize) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU,
				  "scache_rnsize", 2, SCACHE_IOP_RNSIZE); }
ZEND_FUNCTION(scache_rnclear) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU,
				  "scache_rnclear", 2, SCACHE_IOP_RNCLEAR); }
ZEND_FUNCTION(scache_rnpop) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU,
				  "scache_rnpop", 2, SCACHE_IOP_RNPOP); }
ZEND_FUNCTION(scache_rnshift) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU,
				  "scache_rnshift", 2, SCACHE_IOP_RNSHIFT); }


/* two-arg iop-wrappers */
ZEND_FUNCTION(scache_set) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "scache_set", 3, SCACHE_IOP_SET); }
ZEND_FUNCTION(scache_add) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "scache_add", 3, SCACHE_IOP_ADD); }
ZEND_FUNCTION(scache_replace) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "scache_replace", 3, SCACHE_IOP_REPLACE); }
ZEND_FUNCTION(scache_coll) {
	sc_iop_coll(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				"scache_coll", SCACHE_IOP_COLL); }

ZEND_FUNCTION(scache_shset) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "scache_shset", 3, SCACHE_IOP_SHSET); }
ZEND_FUNCTION(scache_shadd) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "scache_shadd", 3, SCACHE_IOP_SHADD); }
ZEND_FUNCTION(scache_shreplace) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "scache_shreplace", 3, SCACHE_IOP_SHREPLACE); }
ZEND_FUNCTION(scache_shcoll) {
	sc_iop_coll(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				"scache_shcoll", SCACHE_IOP_SHCOLL); }

ZEND_FUNCTION(scache_chput) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "scache_chput", 3, SCACHE_IOP_CHPUT); }

ZEND_FUNCTION(scache_rnpush) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU,
				  "scache_rnpush", 3, SCACHE_IOP_RNPUSH); }
ZEND_FUNCTION(scache_rnunshift) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU,
				  "scache_rnunshift", 3, SCACHE_IOP_RNUNSHIFT); }

/* lval-arg */
ZEND_FUNCTION(scache_vadd) {
	sc_iop_lval(INTERNAL_FUNCTION_PARAM_PASSTHRU,
				"scache_vadd", SCACHE_IOP_VADD); }
ZEND_FUNCTION(scache_vsub) {
	sc_iop_lval(INTERNAL_FUNCTION_PARAM_PASSTHRU,
				"scache_vsub", SCACHE_IOP_VSUB); }
ZEND_FUNCTION(scache_vset) {
	sc_iop_lval(INTERNAL_FUNCTION_PARAM_PASSTHRU,
				"scache_vset", SCACHE_IOP_VSET); }

ZEND_FUNCTION(scache_rnrotf) {
	sc_iop_lval(INTERNAL_FUNCTION_PARAM_PASSTHRU,
				"scache_rnrotf", SCACHE_IOP_RNROTF); }
ZEND_FUNCTION(scache_rnrotb) {
	sc_iop_lval(INTERNAL_FUNCTION_PARAM_PASSTHRU,
				"scache_rnrotb", SCACHE_IOP_RNROTB); }

/* ===========================================
   CLASS INTERFACE
*/
static void
sc_class_struct_destroy(void *object TSRMLS_CC) {

	struct sc_classdata_t *classdata =
		(struct sc_classdata_t *)object;

	if (classdata->session) {
		sc_session_clear(classdata->session);
	}

	sc_efree(object);
}

static zend_object_value
sc_class_struct_create(zend_class_entry *ct TSRMLS_DC) {

#if PHP_VERSION_ID < 50399	
	zval *zv;
#endif

	zend_object_value rv;
	struct sc_classdata_t *classdata;

	classdata = (struct sc_classdata_t *)
		sc_emalloc(sizeof(struct sc_classdata_t));
	memset(classdata, 0, 
		   sizeof(struct sc_classdata_t));

	/* classprop */
	zend_object_std_init(&(classdata->std), ct TSRMLS_CC);
#if PHP_VERSION_ID < 50399	
	zend_hash_copy(classdata->std.properties,
				   &(ct->default_properties),
				   (copy_ctor_func_t)zval_add_ref,
				   (void *)&zv, sizeof(zv));
#else
	object_properties_init(&(classdata->std), ct);
#endif
				   
	/* destructor */
	rv.handle = zend_objects_store_put
		(classdata, 
		 (zend_objects_store_dtor_t)zend_objects_destroy_object,
		 sc_class_struct_destroy, NULL TSRMLS_CC);
	rv.handlers = zend_get_std_object_handlers();

	return rv;
}

ZEND_METHOD(SCache, __construct) { 
	
	int host_len = 0, tag_len = 0, sec_len = 0;
	long persistent = 0, port = 0, index = 0, expire = 0;
	char *host = NULL, *tag = NULL, *secret = NULL;

	struct sc_classdata_t *cdata;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
							  "s|sllsll",
							  &tag, &tag_len,
							  &host, &host_len,
							  &expire,
							  &persistent,
							  &secret, &sec_len,
							  &port, &index) == FAILURE) {

		/* invalid parameters */
		zend_throw_exception(scache_ce_exception,
							 "invalid parameters, object creation failed",
							 SCERR_INVALID_REQUEST TSRMLS_CC);
	}

	/* param-verify */
	if (expire == 0) expire = SCACHE_PTO_DEF_EXPIRE;
	if (expire < SCACHE_PTO_MIN_EXPIRE) {
		zend_error(E_WARNING, 
				   "Given expire time %li seconds is below protocol "
				   "minimum. Using protocol minumum %i seconds",
				   expire, SCACHE_PTO_MIN_EXPIRE);
		expire = SCACHE_PTO_MIN_EXPIRE;
	}
	if (expire > SCACHE_PTO_MAX_EXPIRE) {
		zend_error(E_WARNING, 
				   "Given expire time %li seconds is over protocol "
				   "maximum. Using protocol maximum %i seconds",
				   expire, SCACHE_PTO_MAX_EXPIRE);
		expire = SCACHE_PTO_MAX_EXPIRE;
	}

	cdata = (struct sc_classdata_t *)
		zend_object_store_get_object(getThis() TSRMLS_CC);
	cdata->expire = expire;
	cdata->session = sc_connect_real(-1, host, tag, secret, port,
									 !!persistent TSRMLS_CC);
	if (!cdata->session) {

		/* connection failed */
		zend_throw_exception(scache_ce_exception,
							 "no connection, object creation failed",
							 SCERR_NOT_CONNECTED TSRMLS_CC);
	}

}

ZEND_METHOD(SCache, open) {
	sc_open_m(INTERNAL_FUNCTION_PARAM_PASSTHRU, 0); }
ZEND_METHOD(SCache, reset) {
	sc_open_m(INTERNAL_FUNCTION_PARAM_PASSTHRU, 1); }
ZEND_METHOD(SCache, rename) {
	sc_rename(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
			  "SCache->rename"); }
ZEND_METHOD(SCache, drop) {
	sc_drop(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
			"SCache->drop"); }

ZEND_METHOD(SCache, getopt) { 
	sc_opt(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
		   "SCache->getopt", SCACHE_IOP_GETOPT); }
ZEND_METHOD(SCache, setopt) { 
	sc_opt(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
		   "SCache->setopt", SCACHE_IOP_SETOPT); }
ZEND_METHOD(SCache, status) { 
	sc_status(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
			  "SCache->status"); }
ZEND_METHOD(SCache, srvstats) {
	sc_srvstats(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				"SCache->srvstats"); }
ZEND_METHOD(SCache, getindex) {
	sc_getindex_m(INTERNAL_FUNCTION_PARAM_PASSTHRU); }
ZEND_METHOD(SCache, lasterr) { 
	sc_lasterr_m(INTERNAL_FUNCTION_PARAM_PASSTHRU); }
ZEND_METHOD(SCache, strerror) { 
	sc_map_strerror_m(INTERNAL_FUNCTION_PARAM_PASSTHRU); }
ZEND_METHOD(SCache, iov) { 
	sc_iov_m(INTERNAL_FUNCTION_PARAM_PASSTHRU); }

/* one-arg iop-wrappers */
ZEND_METHOD(SCache, get) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "SCache->get", 2, SCACHE_IOP_GET); }
ZEND_METHOD(SCache, clear) { 
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "SCache->clear", 2, SCACHE_IOP_UNSET); }
ZEND_METHOD(SCache, stat) { 
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "SCache->stat", 2, SCACHE_IOP_STAT); }
ZEND_METHOD(SCache, exists) {
	sc_exists_wrapper(INTERNAL_FUNCTION_PARAM_PASSTHRU,
					  "SCache->exists", SCACHE_IOP_STAT); }

ZEND_METHOD(SCache, shget) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "SCache->shget", 2, SCACHE_IOP_SHGET); }
ZEND_METHOD(SCache, shclear) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "SCache->shclear", 2, SCACHE_IOP_SHUNSET); }
ZEND_METHOD(SCache, shstat) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "SCache->shstat", 2, SCACHE_IOP_SHSTAT); }
ZEND_METHOD(SCache, shexists) {
	sc_exists_wrapper(INTERNAL_FUNCTION_PARAM_PASSTHRU,
					  "SCache->shexists", SCACHE_IOP_SHSTAT); }

ZEND_METHOD(SCache, chget) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "SCache->chget", 2, SCACHE_IOP_CHGET); }
ZEND_METHOD(SCache, chclear) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "SCache->chclear", 2, SCACHE_IOP_CHCLEAR); }

ZEND_METHOD(SCache, vget) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU,
				  "SCache->vget", 2, SCACHE_IOP_VGET); }
ZEND_METHOD(SCache, vclear) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU,
				  "SCache->vclear", 2, SCACHE_IOP_VUNSET); }

ZEND_METHOD(SCache, rnget) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU,
				  "SCache->rnget", 2, SCACHE_IOP_RNGET); }
ZEND_METHOD(SCache, rnsize) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU,
				  "SCache->rnsize", 2, SCACHE_IOP_RNSIZE); }
ZEND_METHOD(SCache, rnclear) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU,
				  "SCache->rnclear", 2, SCACHE_IOP_RNCLEAR); }
ZEND_METHOD(SCache, rnpop) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "SCache->rnpop", 2, SCACHE_IOP_RNPOP); }
ZEND_METHOD(SCache, rnshift) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "SCache->rnshift", 2, SCACHE_IOP_RNSHIFT); }


/* two-arg iop-wrappers */
ZEND_METHOD(SCache, set) { 
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "SCache->set", 3, SCACHE_IOP_SET); }
ZEND_METHOD(SCache, add) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "SCache->add", 3, SCACHE_IOP_ADD); }
ZEND_METHOD(SCache, replace) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "SCache->replace", 3, SCACHE_IOP_REPLACE); }
ZEND_METHOD(SCache, coll) {
	sc_iop_coll(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				"SCache->coll", SCACHE_IOP_COLL); }

ZEND_METHOD(SCache, shset) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "SCache->shset", 3, SCACHE_IOP_SHSET); }
ZEND_METHOD(SCache, shadd) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "SCache->shadd", 3, SCACHE_IOP_SHADD); }
ZEND_METHOD(SCache, shreplace) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "SCache->shreplace", 3, SCACHE_IOP_SHREPLACE); }
ZEND_METHOD(SCache, shcoll) {
	sc_iop_coll(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				"SCache->shcoll", SCACHE_IOP_SHCOLL); }

ZEND_METHOD(SCache, chput) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "SCache->chput", 3, SCACHE_IOP_CHPUT); }

ZEND_METHOD(SCache, rnpush) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "SCache->rnpush", 3, SCACHE_IOP_RNPUSH); }
ZEND_METHOD(SCache, rnunshift) {
	sc_iop_single(INTERNAL_FUNCTION_PARAM_PASSTHRU, 
				  "SCache->rnunshift", 3, SCACHE_IOP_RNUNSHIFT); }


/* lval-arg */
ZEND_METHOD(SCache, vadd) {
	sc_iop_lval(INTERNAL_FUNCTION_PARAM_PASSTHRU,
				"SCache->vadd", SCACHE_IOP_VADD); }
ZEND_METHOD(SCache, vsub) {
	sc_iop_lval(INTERNAL_FUNCTION_PARAM_PASSTHRU,
				"SCache->vsub", SCACHE_IOP_VSUB); }
ZEND_METHOD(SCache, vset) {
	sc_iop_lval(INTERNAL_FUNCTION_PARAM_PASSTHRU,
				"SCache->vset", SCACHE_IOP_VSET); }

ZEND_METHOD(SCache, rnrotf) {
	sc_iop_lval(INTERNAL_FUNCTION_PARAM_PASSTHRU,
				"SCache->rnrotf", SCACHE_IOP_RNROTF); }
ZEND_METHOD(SCache, rnrotb) {
	sc_iop_lval(INTERNAL_FUNCTION_PARAM_PASSTHRU,
				"SCache->rnrotb", SCACHE_IOP_RNROTB); }


ZEND_MINIT_FUNCTION(scache) {

	int i;
	zend_class_entry ce, ee; /* class-entry, exception-entry */

	/* **************
	 *  
	 *  can't loop, macro gets sizeof(..[].name) */
	struct {
		int value;
		char *name;
	} reg_const[] = {

		/* iov -function ids */
		{ SCACHE_IOP_GET, "SCIOP_GET" },
		{ SCACHE_IOP_ADD, "SCIOP_ADD" },
		{ SCACHE_IOP_SET, "SCIOP_SET" },
		{ SCACHE_IOP_STAT, "SCIOP_STAT" },
		{ SCACHE_IOP_UNSET, "SCIOP_UNSET" },
		{ SCACHE_IOP_UNSET, "SCIOP_CLEAR" },
		{ SCACHE_IOP_REPLACE, "SCIOP_REPLACE" },
		{ SCACHE_IOP_COLL, "SCIOP_COLL" },
		
		{ SCACHE_IOP_SHGET, "SCIOP_SHGET" },
		{ SCACHE_IOP_SHADD, "SCIOP_SHADD" },
		{ SCACHE_IOP_SHSET, "SCIOP_SHSET" },
		{ SCACHE_IOP_SHSTAT, "SCIOP_SHSTAT" },
		{ SCACHE_IOP_SHUNSET, "SCIOP_SHUNSET" },
		{ SCACHE_IOP_SHUNSET, "SCIOP_SHCLEAR" },
		{ SCACHE_IOP_SHREPLACE, "SCIOP_SHREPLACE" },
		{ SCACHE_IOP_SHCOLL, "SCIOP_SHCOLL" },

		{ SCACHE_IOP_CHGET, "SCIOP_CHGET" },
		{ SCACHE_IOP_CHPUT, "SCIOP_CHPUT" },
		{ SCACHE_IOP_CHCLEAR, "SCIOP_CHCLEAR" },

		{ SCACHE_IOP_VADD, "SCIOP_VADD" },
		{ SCACHE_IOP_VSUB, "SCIOP_VSUB" },
		{ SCACHE_IOP_VGET, "SCIOP_VGET" },
		{ SCACHE_IOP_VSET, "SCIOP_VSET" },
		{ SCACHE_IOP_VUNSET, "SCIOP_VUNSET" },
		{ SCACHE_IOP_VUNSET, "SCIOP_VCLEAR" },
		{ SCACHE_IOP_VBRK, "SCIOP_VBRK" },

		{ SCACHE_IOP_RENAME, "SCIOP_RENAME" },

		{ SCACHE_IOP_RNPUSH, "SCIOP_RNPUSH" },
		{ SCACHE_IOP_RNPOP, "SCIOP_RNPOP" },
		{ SCACHE_IOP_RNSHIFT, "SCIOP_RNSHIFT" },
		{ SCACHE_IOP_RNUNSHIFT, "SCIOP_RNUNSHIFT" },
		{ SCACHE_IOP_RNGET, "SCIOP_RNGET" },
		{ SCACHE_IOP_RNROTF, "SCIOP_RNROTF" },
		{ SCACHE_IOP_RNROTB, "SCIOP_RNROTB" },
		{ SCACHE_IOP_RNSIZE, "SCIOP_RNSIZE" },
		{ SCACHE_IOP_RNCLEAR, "SCIOP_RNCLEAR" },

		{ SCACHE_IOP_LASTERR, "SCIOP_LASTERR" },

		{ SCACHE_QFLAG_OVERWRITE, "SCACHE_QFLAG_OVERWRITE" },
		
		{ SCACHE_OPT_EXPIRES, "SCACHE_OPT_EXPIRES" },
		{ SCACHE_OPT_KEEPONELIMIT, "SCACHE_OPT_KEEPONELIMIT" },

		/* nodetypes */
		{ SCNODE_EMPTY, "SCNODE_EMPTY" },
		{ SCNODE_NONEXISTENT, "SCNODE_NONEXISTENT" },
		{ SCNODE_BRANCH, "SCNODE_BRANCH" },
		{ SCNODE_VALUE, "SCNODE_VALUE" },

		/* errors */
		{ SCERR_SUCCESS, "SCERR_SUCCESS" },
		{ SCERR_PROTOCOL, "SCERR_PROTOCOL" },
		{ SCERR_EXIST, "SCERR_EXIST" },
		{ SCERR_NOTEXIST, "SCERR_NOTEXIST" },
		{ SCERR_SYSTEM, "SCERR_SYSTEM" },
		{ SCERR_INVALID_REQUEST, "SCERR_INVALID_REQUEST" },
		{ SCERR_UNKNOWN_REQUEST, "SCERR_UNKNOWN_REQUEST" },
		{ SCERR_LIMITS_REACHED, "SCERR_LIMITS_REACHED" },
		{ SCERR_NOT_CONNECTED, "SCERR_NOT_CONNECTED" },
		{ SCERR_NO_SESSION, "SCERR_NO_SESSION" },
		{ SCERR_NO_ACCESS, "SCERR_NO_ACCESS" },
		{ SCERR_BREAK, "SCERR_BREAK" },
		{ SCERR_INVALID_FORMAT, "SCERR_INVALID_FORMAT" }
	};

	REGISTER_INI_ENTRIES();
	ZEND_INIT_MODULE_GLOBALS(scache, php_scache_init_globals, NULL);

	/* environment */
#ifndef IOV_MAX
# ifdef _SC_IOV_MAX
    max_iov = sysconf(_SC_IOV_MAX);
# else
    max_iov = 8; /* local default */
# endif
#else
    max_iov = DEFAULT_MAX_IOV;
#endif /* IOV_MAX */

	/* collect static ini entries */
	sc_ini_default_port_changed_i();
	sc_ini_default_host_changed_i();
	sc_ini_conn_timeout_changed_i();
	sc_ini_tag_suffix_changed_i();
	sc_ini_auth_string_changed_i();

	/* register scache connection resource */
	le_sessions =
		zend_register_list_destructors_ex(sc_session_dtor,
										  NULL, SC_SESS_RESNAME,
										  module_number);

	/* register constants */
	for(i = 0; i < (sizeof(reg_const) / sizeof(reg_const[0])); i++) {
		
		/* cannot use macro, macro uses sizeof(reg_const[i].name) */
		zend_register_long_constant(reg_const[i].name,
									strlen(reg_const[i].name) + 1,
									reg_const[i].value,
									CONST_CS | CONST_PERSISTENT,
									module_number);
	}

	/* class interface */
	INIT_CLASS_ENTRY(ce, "SCache", scache_methods);
	INIT_CLASS_ENTRY(ee, "SCacheException", NULL);

	ce.create_object = sc_class_struct_create;

	scache_ce_scache = zend_register_internal_class(&ce TSRMLS_CC);
	scache_ce_exception = zend_register_internal_class_ex
		(&ee, (zend_class_entry *)zend_exception_get_default(TSRMLS_C),
		 NULL TSRMLS_CC);

	/* register session-handler */
#ifdef HAVE_PHP_SESSION
	php_session_register_module(&ps_mod_scache);
#endif

	return SUCCESS;
}


ZEND_MSHUTDOWN_FUNCTION(scache) {

	UNREGISTER_INI_ENTRIES();

	sc_conn_close_all(TSRMLS_C);
	return SUCCESS;
}


ZEND_RINIT_FUNCTION(scache) {
	return SUCCESS;
}

ZEND_RSHUTDOWN_FUNCTION(scache) {
	return SUCCESS;
}

ZEND_MINFO_FUNCTION(scache)
{

	char buffer1[16], buffer2[16], buffer3[16];
	sprintf(buffer1, "%i", SCACHE_PTO_MAX_KEYLEN);
	sprintf(buffer2, "%i", SCACHE_PTO_MAX_VALUELEN);
	sprintf(buffer3, "%i", SCACHE_PTO_MAX_REQLEN);

	php_info_print_table_start();
	php_info_print_table_header(2, "scache support", "enabled");
	php_info_print_table_row(2, "Limits (MAX_KEYLEN)", buffer1);
	php_info_print_table_row(2, "Limits (MAX_VALUELEN)", buffer2);
	php_info_print_table_row(2, "Limits (MAX_REQLEN)", buffer3);
	php_info_print_table_row(2, "Default sock", SCACHED_SOCK);

	php_info_print_box_start(0); /* FIXME */
	PUTS("20130920 SCache version 0.99.3 from "
		 "<a href=\"http://scache.nanona.fi/\">"
		 "http://scache.nanona.fi/</a>");
	php_info_print_box_end();

	php_info_print_table_end();

	DISPLAY_INI_ENTRIES();
}


/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
