<?php

  /* for correct performance measurement you should run
   couple simultaneous tests and combine theirs results.
   otherwise latencies keep scached idling without 
   nothing to do */

define('TEST_ID', uniqid());
define('TEST_PREFIX', 'scache-test-' . TEST_ID);
define('TEST_HOST', '127.0.0.1');
// define('TEST_HOST', 'some.remote.host');

define('LARGEVAL',
       'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
       'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
       'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
       'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
       'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
       'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
       'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
       'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
       'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
       'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
       'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
       'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
       'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
       'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
       'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa');

/* common functions and logging */
function utime_float() {
    list($usec, $sec) = explode(" ", microtime());
    return ((float)$usec + (float)$sec);
}

function m($line, $indent = 1) {
    echo str_repeat('.', ($indent * 4)) . "$line\n";
}

function f($line) {
    m("failed: $line", 2);
}

function run_iov_vvals($conn) {

    $get = Array();
    $add = Array();
    $count = 20;
    for($i = 0, $b = (rand() % 90); $i < $count; $i++) {
	$key = '/vvals/' . ($b + $i) . '/' . ($b + $i);
	$get[$i] = Array(SCIOP_VGET, $key);
	$add[$i] = Array(SCIOP_VADD, $key, 1);
    }

    $ops = array_merge($get, $add, $get);
    if (!($val = scache_iov($conn, $ops))) {
	f("run_iov_vval(): scache_iov(1) failed: " . 
	  scache_strerror(scache_lasterr($conn)));
	die();
    }

    for($i = 0; $i < $count; $i++) {

	$prev = (int)$val[$i];
	$curr = (int)$val[$i + (2 * $count)];

	if (($prev + 1) != $curr) {
	    f("run_iov_vval(): mismatch: $prev + 1 != $curr");
	    die();
	}
    }

    return ($count * 3);
}

function run_iov_shstore($conn) {

    $set = Array();
    $get = Array();
    $count = 20;
    for($i = 0; $i < $count; $i++) {
	$key = '/shared/' .
	    (rand() % 100) . '/' .
	    (rand() % 100);
	$set[$i] = Array(SCIOP_SHSET, $key, $key);
	$get[$i] = Array(SCIOP_SHGET, $key);
    }

    $ops = array_merge($set, $get);
    if (!($val = scache_iov($conn, $ops))) {
	f("run_iov_shstore(): scache_iov(1) failed: " . 
	  scache_strerror(scache_lasterr($conn)));
	die();
    }

    for($i = $count; $i < (2 * $count); $i++) {
	if ($ops[$i][1] != $val[$i]) {
	    f("run_iov_shstore(): mismatch: " . 
	      $val[$i] . "!=" . $ops[$i][1]);
	    die();
	}
    }

    return ($count * 2);
}

function run_iov_sess($conn) {

    $set = Array();
    $get = Array();
    $clr = Array();
    $count = 20;
    for($i = 0; $i < $count; $i++) {
	$key = "$i/$i/$i/$i/$i/$i/$i/$i/$i/$i";
	$set[$i] = Array(SCIOP_SET, $key, $key);
	$get[$i] = Array(SCIOP_GET, $key);
	$clr[$i] = Array(SCIOP_UNSET, $key);
    }
    
    $ops = array_merge($set, $get, $clr);
    if (!($val = scache_iov($conn, $ops))) {
	f("run_iov_sess(): scache_iov(1) failed: " . 
	  scache_strerror(scache_lasterr($conn)));
	die();
    }

    for($i = $count; $i < (2 * $count); $i++) {
	if ($ops[$i][1] != $val[$i]) {
	    f("run_iov_sess(): mismatch: " . 
	      $val[$i] . "!=" . $ops[$i][1]);
	    die();
	}
    }

    return ($count * 3);
}

function run_iov_cache($conn) {

    $ops = Array();
    for($i = 0; $i < 20; $i++) {
	$ops[$i] = 
	    Array(SCIOP_CHGET, '/cache/' .
		  (rand() % 10) . '/' .
		  (rand() % 10) . '/' .
		  (rand() % 10) . '/' .
		  (rand() % 10));
    }

    if (!($val = scache_iov($conn, $ops))) {
	f("run_iov_cache(): scache_iov(1) failed: " . 
	  scache_strerror(scache_lasterr($conn)));
	die();
    }

    for($i = 0; $i < 20; $i++) {
	if ($val[$i]) {
	    $ops[$i][0] = SCIOP_CHCLEAR;
	    if ($val[$i] != $ops[$i][1]) {
		f("run_iov_cache(): mismatch: " . 
		  $val[$i] . "!=" . $ops[$i][1]);
		die();
	    }
	} else {
	    $ops[$i][0] = SCIOP_CHPUT;
	    $ops[$i][2] = $ops[$i][1];
	}
    }

    for($i = 20; $i < 30; $i++) {
	$ops[$i] =
	    Array(SCIOP_CHPUT, "/lrg$i", LARGEVAL);
    }

    if (!scache_iov($conn, $ops)) {
	f("run_iov_cache(): scache_iov(2) failed: " . 
	  scache_strerror(scache_lasterr($conn)));
	die();
    }

    return 50;
}

function run_tests() {

    echo "\n\nRunning scache_iov performance tests\n\n";
    echo "Tests will be run until terminated by Ctrl+C.\n\n";
    echo "If testing locally on multi processor machine, you should have\n";
    echo "couple simultaneous iov-perf-test.phps running and combine\n";
    echo "results.\n\n";
    echo "If testing on a high latency environment (ie. over network)\n";
    echo "you probably need plenty of iov-perf-test.phps running to\n";
    echo "keep scached properly fed.\n\n";
    echo "Expressed results are not cumulative.\n\n";
    echo "Running... please wait:\n\n";

    $conn = scache_reset(TEST_PREFIX, TEST_HOST);
    $count = $i = 0;
    $start = utime_float();
    while(true) {

	$i++;
	$count += run_iov_vvals($conn);
	$count += run_iov_shstore($conn);
	$count += run_iov_sess($conn);
	$count += run_iov_cache($conn);

	if (!($i % 3000)) {

	    $duration = utime_float() - $start;
	    $keys = round($count * 2 / $duration, 0);
	    m("ok, $keys ops/sec");
	    
	    $start = utime_float();
	    $count = 0;
	}
    }
}


run_tests();

/*
  Local variables: ***
  mode: php ***
  c-basic-offset: 4 ***
  End: ***
*/
